from __future__ import annotations

from typing import Any

from labster.di import injector
from labster.domain2.model.profile import Profile
from labster.domain2.model.type_structure import CA, CO, DE, DF, DU, ED, EQ, \
    FA, GRC, IN, LA, LX, SF, UN
from labster.domain2.services.contacts import ContactService
from labster.domain2.services.roles import Role, RoleService

REC = object()
CON = object()
PI = object()

EMPTY_SET = set()

MAP: dict[Any, set[Any]] = {
    UN: {REC},
    FA: {REC},
    DU: {REC},
    CO: {REC, CON, PI},
    SF: {REC},
    DF: {REC},
    LA: {REC, CON, PI},
    ED: {REC, CON, PI},
    DE: EMPTY_SET,
    EQ: EMPTY_SET,
    CA: {REC, CON, PI},
    LX: {REC, CON, PI},
    IN: {REC, CON, PI},
    GRC: {REC, CON, PI},
}


def get_demande_types_for_user(user: Profile) -> set[str]:
    role_service = injector.get(RoleService)
    contact_service = injector.get(ContactService)

    roles = role_service.get_roles_for_user(user)
    structures = roles.get(Role.PORTEUR, set()) | roles.get(Role.GESTIONNAIRE, set())

    # Cf. https://trello.com/c/UM96aY5s
    result = {"faq"}

    for s in structures:
        if not contact_service.get_mapping_for(s):
            continue

        if role_service.has_role(user, Role.PORTEUR, s):
            result |= {"autre"}

        x = MAP[s.type]
        if REC in x:
            result |= {"rh"}
        if PI in x:
            result |= {"pi_invention", "pi_logiciel"}
        if CON in x:
            result |= {"convention", "avenant_convention"}

    return result



"""Stateless role service.

Dummy implementation for now.
"""
from __future__ import annotations

from typing import TYPE_CHECKING, Any

from labster.extensions import db

from . import dgrtt as dgrtt_service

if TYPE_CHECKING:
    from labster.domain.models.profiles import Profile
    from labster.domain.models.roles import Role, RoleType


ROLES_RECHERCHE = ["recherche", "porteur", "gestionnaire", "directeur", "all"]
ROLES_DGRTT = [
    "dgrtt",
    "contact dgrtt",
    "référent",
    "chef de bureau",
    "direction dgrtt",
    "alc",
    "gestionnaire faq",
]
ROLES = ROLES_DGRTT + ROLES_RECHERCHE + ["gouvernance"]


def has_role(user: Profile, role: str, target=None) -> bool:
    """Returns True if user has given role on given target."""
    from labster.domain.models.demandes import Demande
    from labster.domain.models.roles import RoleType

    assert role in ROLES

    # XXX: hack for unit tests
    __roles = getattr(user, "__roles", [])
    if __roles and role in __roles:
        return True

    if role == "alc" and user.uid == "poulainm":
        return True

    ROLES_TO_ROLE_TYPES = {
        "directeur": RoleType.DIRECTION,
        "porteur": RoleType.PORTEUR,
        "contact dgrtt": RoleType.CONTACT_DGRTT,
        "gestionnaire faq": RoleType.ALC,  # TODO
        "alc": RoleType.ALC,
        "all": RoleType.ALL,
    }
    if role in ROLES_TO_ROLE_TYPES:
        return _has_role(user, ROLES_TO_ROLE_TYPES[role], target)

    if role == "recherche":
        return user.laboratoire is not None

    if role == "gouvernance":
        return bool(user.gouvernance_vraiment)

    if role == "dgrtt":
        return bool(user.dgrtt)

    # Rôles "recherche"
    if role == "gestionnaire":
        if isinstance(target, Demande):
            return has_role(user, role, target.structure)

        if _has_role(user, RoleType.GDL, target):
            return True
        if not target:
            return False
        parent = target.parent
        if not parent:
            return False
        if _has_role(user, RoleType.GDL, parent):
            return True

        parent = parent.parent
        if not parent:
            return False
        return _has_role(user, RoleType.GDL, parent)

    if target:
        return has_contextual_role(user, role, target)

    # Rôles DGRTT
    if role == "référent":
        return dgrtt_service.est_referent(user)

    if role == "chef de bureau":
        return bool(user.chef_du_bureau)

    if role == "direction dgrtt":
        fsp = user.fonction_structurelle_principale
        return (
            has_role(user, "dgrtt") and fsp and "Directeur de services centraux" in fsp
        )

    if role == "admin":
        raise RuntimeError("Role 'admin' has been replaced by 'alc'")

    raise NotImplementedError(f"Le rôle '{role}' n'existe pas")


def has_contextual_role(user: Profile, role: str, target) -> bool:
    assert role in ROLES

    from labster.domain.models.demandes import Demande
    from labster.domain.models.unites import OrgUnit

    # TODO: forward to OrgUnit instead ?
    if isinstance(target, OrgUnit):
        return has_role(user, role) and user in target.membres

    if isinstance(target, Demande):
        return has_role(user, role, target.structure)

    return False


def rename_role_dgrtt_to_dri(role):
    """Simple rename instead of data migration."""
    if "dgrtt" in role.lower():
        role = role.replace("DGRTT", "DR&I")
        role = role.replace("dgrtt", "DR&I")
    return role


def all_roles(user):
    roles = []
    for role in ROLES:
        if has_role(user, role):
            roles.append(rename_role_dgrtt_to_dri(role))
    return roles


#
# DB-backed roles
#
def _has_role(user, role_type, target=None):
    # type: (Profile, RoleType, Any) -> bool

    from labster.domain.models.roles import RoleType

    assert isinstance(role_type, RoleType)

    roles = user.roles
    if target:
        return any(r.type == role_type.value for r in roles if r.context == target)
    else:
        return any(r.type == role_type.value for r in roles)


def get_roles(user=None, role_type=None, target=None) -> list[Role]:
    from labster.domain.models.roles import Role, RoleType

    assert isinstance(role_type, (type(None), RoleType))
    assert user or role_type or target

    query = Role.query
    if user:
        query = query.filter_by(profile=user)
    if role_type:
        query = query.filter_by(type=role_type.value)
    if target:
        query = query.filter_by(context=target)
    return query.all()


def set_role_value(user, role: str, value: bool, target=None):
    from labster.domain.models.roles import RoleType

    assert isinstance(role, str)

    mapping = {
        # Rôles labos
        "porteur": RoleType.PORTEUR,
        "directeur": RoleType.DIRECTION,
        "gestionnaire": RoleType.GDL,
        "all": RoleType.ALL,
        # Rôles DGRTT
        "contact dgrtt": RoleType.CONTACT_DGRTT,
        "alc": RoleType.ALC,
    }
    try:
        role_type = mapping[role]
    except KeyError:
        raise RuntimeError(f"Unknown role: {role}")

    if value and not _has_role(user, role_type, target):
        grant_role(user, role_type, target)

    elif not value and _has_role(user, role_type, target):
        ungrant_role(user, role_type, target)


def grant_role(user: Profile, role_type: RoleType, target=None):
    from labster.domain.models.roles import Role, RoleType

    assert isinstance(role_type, RoleType)

    roles = (
        Role.query.filter(Role.profile == user)
        .filter(Role.type == role_type.value)
        .filter(Role.context == target)
        .all()
    )
    assert len(roles) == 0

    role = Role(type=role_type.value, profile=user, context=target)
    db.session.add(role)


def ungrant_role(user: Profile, role_type: RoleType, target=None):
    from labster.domain.models.roles import Role, RoleType

    assert isinstance(role_type, RoleType)

    roles = (
        Role.query.filter(Role.profile == user)
        .filter(Role.type == role_type.value)
        .filter(Role.context == target)
        .all()
    )
    assert len(roles) == 1
    role = roles[0]
    assert role in user.roles
    index = user.roles.index(role)
    del user.roles[index]


# Utilisé en phase de test (URL: /switch/).
def get_all_users() -> list[Profile]:
    from labster.domain.models.profiles import Profile

    gestionnaires = {"pulcherie", "boyern", "courtoisi", "sos", "girardv"}
    porteurs = {"carapezzi", "duhieu", "lombard", "diasdeamorim", "valdes"}
    directeurs = {
        "santiardbaro",
        "charretier",
        "sciandra",
        "mercierc",
        "mouchelj",
        "stemmann",
    }

    all_uids = gestionnaires | porteurs | directeurs

    users = Profile.query.filter(Profile.uid.in_(all_uids)).all()

    users += list(dgrtt_service.get_membres())
    users = sorted(users, key=lambda x: x.uid)

    return users



"""Modèle de case management adapté au projet Labster."""
from __future__ import annotations

from datetime import date, datetime
from typing import Container, Dict

import structlog

from labster.domain.services.notifications import send_email
from labster.forms.workflow import ConfirmerFinalisationForm, \
    ConfirmerRecevabiliteForm, WorkflowForm
from labster.lib.workflow import State, Transition, Workflow

logger = structlog.get_logger()


class EnEdition(State):
    label = "En édition"
    next_action = "Edition à finaliser et à soumettre"

    def on_enter(self, workflow):
        case = workflow.case
        case.active = True
        case.editable = True

    def task_owners(self, workflow):
        case = workflow.case
        return {u for u in [case.gestionnaire, case.porteur] if u}


class EnValidation(State):
    label = "En cours de validation hiérarchique"
    label_short = "En validation"
    next_action = "Demande à considérer pour validation"

    def task_owners(self, workflow):
        demande = workflow.case
        # assert validation_stage
        if not demande.wf_stage:
            logger.warning(
                f"Warning: la demande {demande.id} n'a pas de validation_stage"
            )
            demande.wf_stage = next_validation_stage(demande)
        return demande.wf_stage.direction


def next_validation_stage(demande):
    from labster.domain.models.unites import DEPARTEMENT, EQUIPE, LABORATOIRE

    structure = demande.structure
    stage = demande.wf_stage

    if not stage:
        stage = structure
        if stage.type == EQUIPE and not structure.wf_must_validate(EQUIPE):
            stage = stage.parent
        if stage.type == DEPARTEMENT and not structure.wf_must_validate(DEPARTEMENT):
            stage = stage.parent
        if stage.type == LABORATOIRE and not structure.wf_must_validate(LABORATOIRE):
            stage = None
        assert stage is None or stage.type in [EQUIPE, DEPARTEMENT, LABORATOIRE]
        return stage

    stage = stage.parent
    if not stage:
        return None

    if stage.type == DEPARTEMENT and not structure.wf_must_validate(DEPARTEMENT):
        stage = stage.parent
    if not stage:
        return None

    if stage.type == LABORATOIRE and not structure.wf_must_validate(LABORATOIRE):
        stage = None
    if not stage:
        return None

    if stage == None or stage.type in [DEPARTEMENT, LABORATOIRE]:
        return stage

    return None


class EnVerification(State):
    label = "Recevabilité en cours de vérification"
    label_short = "En vérification"
    next_action = "Recevabilité à confirmer"

    def on_enter(self, workflow):
        workflow.set_value("ar_envoye", False)

    def task_owners(self, workflow):
        case = workflow.case
        if case.contact_dgrtt:
            return [case.contact_dgrtt]
        else:
            return []


class EnInstruction(State):
    label = "En cours d'instruction par la DR&I"
    label_short = "En instruction"
    next_action = "Instruction à mener et finaliser"

    def task_owners(self, workflow):
        case = workflow.case
        if case.contact_dgrtt:
            return [case.contact_dgrtt]
        else:
            return []


# Etats finaux
class Traitee(State):
    label = "Traitée par la DR&I"
    label_short = "Traitée"
    is_final = True


class Rejetee(State):
    label = "Rejetée par la DR&I"
    label_short = "Rejetée"
    is_final = True


class Abandonnee(State):
    label = "Abandonnée par le porteur"
    label_short = "Abandonnée"
    is_final = True


EN_EDITION = EnEdition()
EN_VALIDATION = EnValidation()
EN_VERIFICATION = EnVerification()
EN_INSTRUCTION = EnInstruction()
TRAITEE = Traitee()
REJETEE = Rejetee()
ABANDONNEE = Abandonnee()

ACTIVE_STATES: list[State] = [
    EN_EDITION,
    EN_VALIDATION,
    EN_VERIFICATION,
    EN_INSTRUCTION,
]
INACTIVE_STATES: list[State] = [TRAITEE, REJETEE, ABANDONNEE]
ALL_STATES: list[State] = ACTIVE_STATES + INACTIVE_STATES


#
# Transitions
#
class Abandonner(Transition):
    label = "Abandonner la demande"
    category = "danger"
    from_states = [EN_EDITION, EN_VALIDATION, EN_VERIFICATION, EN_INSTRUCTION]
    to_state = ABANDONNEE
    message = "{actor} a abandonné la demande."

    def precondition(self, workflow):
        return workflow.actor_is_porteur_or_gdl()

    def get_users_to_notify(self, workflow, old_state):
        case = workflow.case
        if old_state == EN_EDITION:
            return []
        if old_state == EN_VALIDATION:
            return case.owners + case.structure.direction
        if old_state in (EN_VERIFICATION, EN_INSTRUCTION):
            return case.owners + [case.contact_dgrtt]
        raise RuntimeError(f"Unknown state: {old_state}")


class Desarchiver(Transition):
    label = "Désarchiver la demande"
    category = "danger"
    from_states = [ABANDONNEE, REJETEE, TRAITEE]
    message = "{actor} a désarchivé la demande."

    def apply(self, workflow, data):
        from labster.domain.models.demandes import Demande

        demande = workflow.case  # type: Demande
        old_state_id = demande.wf_history[-1].get("old_state", "EN_EDITION")
        old_state = workflow.get_state_by_id(old_state_id)
        old_state.enter(workflow)
        demande.active = True
        demande.editable = True

    def get_form(self, workflow, **kw):
        return WorkflowForm(require_note=True)

    def get_users_to_notify(self, workflow, old_state):
        case = workflow.case
        return case.owners


class Soumettre(Transition):
    label = "Soumettre la demande"
    from_states = [EN_EDITION]

    def precondition(self, workflow):
        demande = workflow.case
        structure = demande.structure

        if not demande.is_valid():
            return False

        if workflow.actor == demande.porteur:
            return True

        # Cas à gérer: la demande est issue d'une sous-structure dans
        # le périmètre du gestionnaire.
        #
        # demande.structure et workflow.actor.structure peuvent être différentes.
        #
        # Par exemple:
        # demande.structure = equipe Developpement des circuits neuronaux
        # workflow.actor.structure = Institut de la vision
        # Et dans ce cas
        # demande.structure.get_gestionnaires() = []
        #
        # Sur le formulaire de la demande, l'utilisateur voit
        # "Institut de la vision", donc peut aller changer son paramètre "permettre_soummission_directe" à True
        # On a donc
        # demande.structure.permettre_soummission_directe inchangé, potentiellement à False,
        # et workflow.actor.structure.permettre_soummission_directe à True.
        #
        # On a donc besoin des deux "if" suivants.
        if (
            workflow.actor.structure
            and workflow.actor.structure.permettre_soummission_directe
            and workflow.actor in workflow.actor.structure.get_gestionnaires()
        ):
            return True

        if (
            structure.permettre_soummission_directe
            and workflow.actor in structure.get_gestionnaires()
        ):
            return True

        if (
            workflow.get_value("validee_hierarchie")
            and structure.permettre_reponse_directe
            and workflow.actor == demande.gestionnaire
        ):
            return True

        return False

    def apply(self, workflow, data):
        demande = workflow.case
        demande.editable = False

        if data.get("resoumission"):
            workflow.set_value("validee_hierarchie", False)

        if workflow.get_value("validee_hierarchie"):
            if workflow.get_value("recevable"):
                EN_INSTRUCTION.enter(workflow)
            else:
                EN_VERIFICATION.enter(workflow)
        else:
            demande.wf_stage = next_validation_stage(demande)
            EN_VALIDATION.enter(workflow)

    def message(self, workflow):
        if workflow.get_value("validee_hierarchie"):
            return "{actor} a resoumis sa demande sans revalidation hiérarchique"
        else:
            return "{actor} a soumis sa demande pour validation hiérarchique."

    def get_users_to_notify(self, workflow, old_state):
        case = workflow.case
        if workflow.get_value("validee_hierarchie"):
            return [case.contact_dgrtt]
        else:
            return case.wf_stage.direction

    def get_form(self, workflow, **kw):
        if workflow.get_value("validee_hierarchie"):
            return WorkflowForm(ask_for_revalidation=True)

        return WorkflowForm()


#
# Prendre la main
#
class AbstractPrendreLaMain(Transition):
    label = "Prendre la main sur la demande"
    category = "danger"
    message = "{actor} a pris la main sur la demande."

    def get_users_to_notify(self, workflow, old_state):
        from labster.domain.models.profiles import Profile

        case = workflow.case
        actor = workflow.actor
        user_ids = {entry["actor_id"] for entry in case.wf_history}
        users = []
        for id in user_ids:
            try:
                user = Profile.query.get(id)
                if user != actor:
                    users.append(user)
            except Exception:
                pass
        return users


class PrendreLaMainGestionnaire(AbstractPrendreLaMain):
    from_states = ACTIVE_STATES

    def precondition(self, workflow):
        case = workflow.case
        actor = workflow.actor
        return actor.has_role("gestionnaire", case) and actor not in (
            case.gestionnaire,
            case.porteur,
        )

    def apply(self, workflow, data):
        case = workflow.case
        actor = workflow.actor
        case.gestionnaire = actor


class PrendreLaMainDgrtt(AbstractPrendreLaMain):
    from_states = ACTIVE_STATES

    def precondition(self, workflow):
        case = workflow.case
        actor = workflow.actor
        return actor.has_role("dgrtt") and not actor == case.contact_dgrtt

    def apply(self, workflow, data):
        case = workflow.case
        actor = workflow.actor
        case.contact_dgrtt = actor


#
# "Valider demande (hiérarchie)
#
class ValiderDir(Transition):
    label = "Valider la demande"
    from_states = [EN_VALIDATION]
    message = "Demande validée par la hiérarchie ({actor})."

    def precondition(self, workflow):
        return workflow.actor in workflow.state.task_owners(workflow)

    def apply(self, workflow, data):
        demande = workflow.case
        demande.assigne_contact_dgrtt()
        workflow.set_value("validee_hierarchie", True)
        demande.date_effective = date.today()

        next_stage = next_validation_stage(demande)
        demande.wf_stage = next_stage
        if not next_stage:
            EN_VERIFICATION.enter(workflow)

    def get_users_to_notify(self, workflow, old_state):
        case = workflow.case
        if case.contact_dgrtt:
            return case.owners + [case.contact_dgrtt]
        else:
            return case.owners


class RequerirModificationDir(Transition):
    label = "Requérir modification / complément"
    from_states = [EN_VALIDATION]
    to_state = EN_EDITION
    message = (
        "Demande de compléments / modifications par {actor} "
        "(direction labo/département/équipe) pour vérification de recevabilité."
    )

    def precondition(self, workflow):
        return workflow.actor in workflow.state.task_owners(workflow)

    def apply(self, workflow, data):
        demande = workflow.case
        demande.wf_stage = None
        workflow.set_value("validee_hierarchie", False)

    def get_users_to_notify(self, workflow, old_state):
        case = workflow.case
        return case.owners

    def get_form(self, workflow, **kw):
        return WorkflowForm(require_note=True)


#
# DGRTT (renommé DR&I)
#
class AccuserReception(Transition):
    label = "Accuser réception en attendant vérification ultérieure"
    from_states = [EN_VERIFICATION]
    to_state = EN_VERIFICATION
    message = "Accusé de réception envoyé par {actor} (contact)."

    def precondition(self, workflow):
        return workflow.actor_is_contact_dgrtt() and not workflow.get_value(
            "ar_envoye", False
        )

    def apply(self, workflow, data):
        workflow.set_value("ar_envoye", True)

    def get_users_to_notify(self, workflow, old_state):
        case = workflow.case
        return case.owners


class RequerirModificationDgrtt(Transition):
    label = "Requérir modification / complément"
    from_states = [EN_VERIFICATION, EN_INSTRUCTION]
    to_state = EN_EDITION

    def precondition(self, workflow):
        return workflow.actor_is_contact_dgrtt()

    def apply(self, workflow, data):
        if data.get("resoumission"):
            workflow.set_value("validee_hierarchie", False)

    def message(self, workflow):
        if workflow.state == EN_VERIFICATION:
            return (
                "Demande de compléments / modifications par {actor} "
                "(contact) pour vérification de recevabilité."
            )
        else:
            return (
                "Demande de compléments / modifications par {actor} "
                "(contact) pour instruction."
            )

    def get_users_to_notify(self, workflow, old_state):
        case = workflow.case
        return case.owners

    def get_form(self, workflow, **kw):
        if workflow.get_value("validee_hierarchie"):
            return WorkflowForm(ask_for_revalidation=True, require_note=True)

        return WorkflowForm(require_note=True)


class ConfirmerRecevabiliteDgrtt(Transition):
    label = "Confirmer recevabilité"
    from_states = [EN_VERIFICATION]
    to_state = EN_INSTRUCTION

    def precondition(self, workflow):
        return workflow.actor_is_contact_dgrtt()

    def apply(self, workflow, data):
        workflow.set_value("recevable", True)
        self.send_notification(workflow)

    def send_notification(self, workflow):
        case = workflow.case
        subject = "Recevabilité de votre demande par la DR&I"
        ctx = {
            "transition": self,
            "demande": case,
            "workflow": workflow,
            "now": datetime.now(),
        }
        send_email(case.owners, subject, "notif-demande-recevable.html", ctx)

    def get_users_to_notify(self, workflow, old_state):
        case = workflow.case
        return case.owners

    def get_form(self, workflow, **kw):
        return ConfirmerRecevabiliteForm()

    def message(self, workflow):
        demande = workflow.case
        tpl = (
            "Recevabilité confirmée par {actor} (contact). "
            "No Infolab: %s." % demande.no_infolab
        )
        return tpl


class ConfirmerFinalisationDgrtt(Transition):
    label = "Confirmer finalisation"
    from_states = [EN_INSTRUCTION]
    to_state = TRAITEE

    def precondition(self, workflow):
        return workflow.actor_is_contact_dgrtt()

    def get_users_to_notify(self, workflow, old_state):
        case = workflow.case
        return case.owners

    def get_form(self, workflow, **kw):
        return ConfirmerFinalisationForm()

    def apply(self, workflow, data):
        self.send_notification(workflow)

    def send_notification(self, workflow):
        case = workflow.case
        ctx = {
            "transition": self,
            "demande": case,
            "workflow": workflow,
            "now": datetime.now(),
        }
        subject = "Finalisation de votre demande par la DR&I"
        du = case.laboratoire.direction
        recipients = case.owners + du
        send_email(recipients, subject, "notif-demande-finalisee.html", ctx)

    def message(self, workflow):
        demande = workflow.case
        tpl = (
            "Traitement finalisé par {actor} (contact)."
            "No eOTP: %s." % demande.no_eotp
        )
        return tpl


class RejeterDgrtt(Transition):
    label = "Rejeter / abandonner demande"
    category = "danger"
    from_states = [EN_EDITION, EN_VALIDATION, EN_VERIFICATION, EN_INSTRUCTION]
    to_state = REJETEE
    message = "Demande rejetée / abandonnées par {actor} (contact)."

    def precondition(self, workflow):
        actor = workflow.actor
        return actor.has_role("dgrtt")

    def get_users_to_notify(self, workflow, old_state):
        case = workflow.case
        return case.owners

    def apply(self, workflow, data):
        # type: (Workflow, Dict) -> None
        self.send_notification(workflow)

    def send_notification(self, workflow):
        # type: (Workflow) -> None
        case = workflow.case
        ctx = {
            "transition": self,
            "demande": case,
            "workflow": workflow,
            "now": datetime.now(),
        }
        subject = "Rejet de votre demande par la DR&I"
        send_email(case.owners, subject, "notif-demande-rejetee.html", ctx)

    def get_form(self, workflow, **kw):
        return WorkflowForm(require_note=True)


class Commenter(Transition):
    label = "Envoyer un message"
    category = "success"
    from_states = [EN_EDITION, EN_VERIFICATION, EN_INSTRUCTION]
    message = "{actor} a posté le commentaire ou la question suivante: "

    def precondition(self, workflow):
        # type: (Workflow) -> bool

        actor = workflow.actor
        demande = workflow.case
        if workflow.state == EN_EDITION and not demande.contact_dgrtt:
            return False
        return actor in self._get_stakeholder(workflow)

    def get_users_to_notify(self, workflow, old_state):
        return self._get_stakeholder(workflow)

    def _get_stakeholder(self, workflow):
        # type: (Workflow) -> Container
        from .profiles import Profile
        from .unites import OrgUnit

        demande = workflow.case
        structure = demande.structure  # type: OrgUnit

        stakeholders = set(demande.owners)
        if demande.contact_dgrtt:
            stakeholders.add(demande.contact_dgrtt)

        directeurs = structure.get_directeurs()
        for directeur in directeurs:
            stakeholders.add(directeur)

        for history_item in demande.wf_history:
            actor_id = history_item["actor_id"]
            if actor_id:
                actor = Profile.query.get(actor_id)
                stakeholders.add(actor)

        return stakeholders

    def get_form(self, workflow, **kw):
        return WorkflowForm(require_note=True)

    def apply(self, workflow, data):
        # type: (Workflow, Dict) -> None
        self.send_notification(workflow)

    def send_notification(self, workflow):
        # type: (Workflow) -> None
        case = workflow.case
        ctx = {
            "transition": self,
            "demande": case,
            "workflow": workflow,
            "now": datetime.now(),
        }
        subject = "Un commentaire sur votre demande"
        send_email(case.owners, subject, "notif-demande-comment.html", ctx)


#
ABANDONNER = Abandonner()
DESARCHIVER = Desarchiver()
SOUMETTRE = Soumettre()
PRENDRE_LA_MAIN_GESTIONNAIRE = PrendreLaMainGestionnaire()
PRENDRE_LA_MAIN_DGRTT = PrendreLaMainDgrtt()
#
VALIDER_DIR = ValiderDir()
REQUERIR_MODIFICATION_DIR = RequerirModificationDir()
#
ACCUSER_RECEPTION = AccuserReception()
REQUERIR_MODIFICATION_DGRTT = RequerirModificationDgrtt()
CONFIRMER_RECEVABILITE_DGRTT = ConfirmerRecevabiliteDgrtt()
CONFIRMER_FINALISATION_DGRTT = ConfirmerFinalisationDgrtt()
REJETER_DGRTT = RejeterDgrtt()
#
COMMENTER = Commenter()


#
# Workflow
#
class LabsterWorkflow(Workflow):
    initial_state = EN_EDITION

    states = ALL_STATES

    # NB: order counts!
    transitions = [
        SOUMETTRE,
        PRENDRE_LA_MAIN_GESTIONNAIRE,
        VALIDER_DIR,
        PRENDRE_LA_MAIN_DGRTT,
        REQUERIR_MODIFICATION_DIR,
        ACCUSER_RECEPTION,
        CONFIRMER_RECEVABILITE_DGRTT,
        CONFIRMER_FINALISATION_DGRTT,
        REQUERIR_MODIFICATION_DGRTT,
        REJETER_DGRTT,
        ABANDONNER,
        DESARCHIVER,
        COMMENTER,
    ]

    def actor_is_contact_dgrtt(self):
        return self.actor == self.case.contact_dgrtt

    def actor_is_porteur_or_gdl(self):
        return self.actor in (self.case.porteur, self.case.gestionnaire)



from __future__ import annotations

from injector import Injector

from labster.domain2.model.structure import Structure, StructureRepository
from labster.domain2.model.type_structure import CO, DU, FA, LA, UN


def test_single():
    universite = Structure(nom="Sorbonne Université", type_name=UN.name, sigle="SU")
    assert universite.nom == "Sorbonne Université"
    assert universite.name == "Sorbonne Université"
    assert universite.sigle_ou_nom == "SU"
    assert universite.is_reelle
    assert universite.active
    assert len(universite.ancestors) == 0
    assert len(universite.descendants) == 0
    universite.check()

    universite.delete()
    assert not universite.active


def test_hierarchy():
    universite = Structure(nom="Sorbonne Université", type_name=UN.name)
    fac_sciences = Structure(nom="Faculté des Sciences", type_name=FA.name)
    assert universite not in fac_sciences.parents
    assert fac_sciences not in universite.children

    universite.add_child(fac_sciences)
    assert universite in fac_sciences.parents
    assert fac_sciences in universite.children
    assert universite.depth == 0
    assert fac_sciences.depth == 1
    assert fac_sciences.ancestors == [universite]
    universite.check()
    fac_sciences.check()

    universite.remove_child(fac_sciences)
    assert universite not in fac_sciences.parents
    assert fac_sciences not in universite.children
    assert universite.depth == 0
    assert fac_sciences.depth == 0
    universite.check()
    fac_sciences.check()

    fac_sciences.add_parent(universite)
    assert universite in fac_sciences.parents
    assert fac_sciences in universite.children
    assert universite.depth == 0
    assert fac_sciences.depth == 1
    universite.check()
    fac_sciences.check()

    fac_sciences.remove_parent(universite)
    assert universite not in fac_sciences.parents
    assert fac_sciences not in universite.children
    assert universite.depth == 0
    assert fac_sciences.depth == 0
    universite.check()
    fac_sciences.check()


def test_deep_hierarchy():
    universite = Structure(nom="Sorbonne Université", type_name=UN.name)
    fac = Structure(nom="Faculté", type_name=FA.name)
    composante = Structure(nom="Composante", type_name=CO.name)
    labo = Structure(nom="Labo", type_name=LA.name)

    universite.add_child(fac)
    fac.add_child(composante)
    composante.add_child(labo)

    universite.check()
    fac.check()
    composante.check()
    labo.check()

    assert labo.ancestors == [composante, fac, universite]


def test_constraints_on_parent():
    un = Structure(nom="Sorbonne Université", type_name=UN.name)
    la = Structure(nom="Labo", type_name=LA.name)
    du = Structure(nom="DU", type_name=DU.name)

    assert not un.can_have_parent(un)
    assert not un.can_have_parent(la)
    assert not la.can_have_parent(la)
    assert not la.can_have_parent(un)

    assert not un.can_have_parent(du)
    assert du.can_have_parent(un)

    assert not un.can_have_child(un)
    assert not un.can_have_child(la)
    assert not la.can_have_child(la)
    assert not la.can_have_child(un)

    assert un.can_have_child(du)
    assert not du.can_have_child(un)


def test_repo(injector: Injector, db_session):
    repo = injector.get(StructureRepository)

    universite = Structure(
        nom="Sorbonne Université", type_name=UN.name, sigle="SU", dn="Top"
    )
    fac_sciences = Structure(nom="Faculté des Sciences", type_name=FA.name)
    repo.put(universite)
    repo.put(fac_sciences)
    assert universite in repo.get_all()
    assert fac_sciences in repo.get_all()
    repo.check_all()

    assert universite == repo.get_by_id(universite.id)
    assert universite == repo.get_by_dn(universite.dn)
    assert universite == repo.get_by_sigle(universite.sigle)

    universite.add_child(fac_sciences)
    assert universite in repo.get_all()
    assert fac_sciences in repo.get_all()
    repo.check_all()



#!/usr/bin/env python

"""Cron-like scheduler."""

import os
import sys

import click
from apscheduler.schedulers.blocking import BlockingScheduler
from apscheduler.triggers.cron import CronTrigger
from loguru import logger

job_defaults = {
    'coalesce': False,
    'max_instances': 4
}
scheduler = BlockingScheduler(job_defaults=job_defaults)

logger.remove(0)
logger.add(sys.stdout, format="{time} {level} {message}")


@click.group()
def cli():
    pass


@cli.command()
def list():
    scheduler.print_jobs()


@cli.command()
def run():
    logger.info("Starting job scheduler")
    scheduler.start()


def system(cmd: str):
    logger.info(cmd)
    status = os.system(cmd)
    logger.info(cmd + " exited with status {}", status)


@scheduler.scheduled_job(CronTrigger(hour=1, minute=0, second=0))
def daily_reindex():
    system("./bin/flask reindex")


@scheduler.scheduled_job(CronTrigger(hour=6, minute=0, second=0))
def daily_sync():
    system("./bin/flask ldap-sync")
    system("./bin/flask syncbi")
    system("./bin/flask update-retard")


@scheduler.scheduled_job(CronTrigger(hour=0, minute=0, second=0))
def daily_recap():
    system("./bin/flask send-notifications daily")


@scheduler.scheduled_job(CronTrigger(day_of_week="thu", hour=0, minute=0, second=0))
def weekly_notif():
    system("./bin/flask send-notifications weekly")


@scheduler.scheduled_job(CronTrigger(day_of_week="tue", hour=0, minute=0, second=0))
def weekly_recap():
    system("./bin/flask send-recap")


if __name__ == "__main__":
    cli()



"""
TODO: explain
"""
from __future__ import annotations

from abilian.core.entities import Entity
from abilian.core.sqlalchemy import JSONDict
from sqlalchemy import Column


class Config(Entity):
    __tablename__ = "config"
    __indexable__ = False

    data = Column(JSONDict(), default=dict)



from __future__ import annotations

from pathlib import Path
from typing import Callable

import pkg_resources
import structlog
from abilian.services import audit_service
from flask import Flask, current_app, g, request, session
from toolz import first, memoize
from werkzeug.utils import import_string

from labster.util import url_for

from . import search
from .bus import register_callback
from .security import get_current_user, login_user

BLUEPRINTS = [
    "labster.rpc",
    "labster.blueprints.v3",
    "labster.blueprints.main",
    "labster.blueprints.auth",
    "labster.blueprints.backup",
    "labster.blueprints.notifications",
]


def init_web(app: Flask) -> None:
    app.context_processor(inject_polymorphic_url_for)

    register_blueprints(app)

    app.before_request(login_user)
    app.before_request(stop_services)
    app.before_request(inject_debug_info)
    app.before_request(make_session_permanent)
    app.before_request(inject_assets)
    app.before_request(setup_cache)

    app.jinja_env.filters.update(datetime=lambda x: x.strftime("%d/%m/%y %H:%M"))

    search.register(app)

    register_callback()


def register_blueprints(app: Flask) -> None:
    logger = structlog.get_logger()

    for fqn in BLUEPRINTS:
        module = import_string(fqn)
        logger.info("Registering blueprint", name=module.__name__)
        app.register_blueprint(module.blueprint)


def make_session_permanent() -> None:
    session.permanent = True


def inject_debug_info() -> None:
    debug_info = {}
    current_user = get_current_user()
    if current_user.is_authenticated:
        debug_info["user_uid"] = current_user.uid
    debug_info["url"] = request.url

    g.debug_info = debug_info

    try:
        ws = pkg_resources.working_set
        version = ws.by_key["labster"]._version  # type: ignore
    except:
        version = "???"

    g.labster_version = version


def inject_assets() -> None:
    @memoize
    def get_assets_from_filesystem():
        root = Path(current_app.root_path)
        static = root / "static"
        css = static / "css"
        js = static / "js"
        start = len(str(root))
        css_assets = [
            str(first(css.glob("chunk*")))[start:],
            str(first(css.glob("app*")))[start:],
        ]
        js_assets = [
            str(first(js.glob("chunk*")))[start:],
            str(first(js.glob("app*")))[start:],
        ]
        return {"css": css_assets, "js": js_assets}

    if current_app.config.get("PRODUCTION"):
        g.assets = get_assets_from_filesystem()
    else:
        g.assets = {"css": [], "js": ["http://labster.local:8080/app.js"]}


#
# Various temporary hacks
#
def stop_services() -> None:
    # TODO later: audit service needs Users, not Profiles
    if audit_service.running:
        audit_service.stop()
        # if index_service.running:
        #     index_service.stop()


def inject_polymorphic_url_for() -> dict[str, Callable]:
    # TODO: use real registry
    return {"url_for": url_for, "id": id}


def setup_cache() -> None:
    g.cache = {}



"""Formulaires PI: parties communes."""
# Nom
# Prénom
# Téléphone
# Mail
# Nationalité / Nationality
# Adresse personnelle (n°, rue, bâtiment, CP, ville, pays) / personal address
# "Part contributive en % / Contributive share - représente le pourcentage
# de participation à la création, qui conditionne le futur intéressement des contributeurs"
# Période à laquelle vous avez contribué à l'œuvre / period in which you contributed to the work
# Indiquer la nature de la contribution à l'œuvre / Please indicate the contribution to the work
from __future__ import annotations

from labster.newforms.base import Boolean2Field, BooleanField, FieldSet, \
    Select2Field, StringField, TextAreaField

interlocuteur_privilegie = FieldSet(
    "interlocuteur_privilegie",
    "Interlocuteur privilégié",
    [
        Select2Field(
            "interlocuteur_privilegie",
            "Interlocuteur privilégié",
            choices=[],
            required=True,
        ),
        StringField("nationalite", "Nationalité / Nationality", required=True),
        TextAreaField(
            "adresse_personnelle",
            "Adresse personnelle (n°, rue, bâtiment, CP, ville, pays) / personal address",
            required=True,
        ),
        StringField(
            "part_contributive",
            "Part contributive en % / Contributive share - <em>représente le pourcentage de participation à la création, qui conditionne le futur intéressement des contributeurs</em>",
            required=True,
        ),
        StringField(
            "periode_contribution",
            "Période à laquelle vous avez contribué à l'œuvre / period in which you contributed to the work",
            required=True,
        ),
        TextAreaField(
            "nature_contribution",
            "Indiquer la nature de la contribution à l'œuvre / Please indicate the contribution to the work",
            required=True,
        ),
    ],
)

porteur_de_projet = FieldSet(  #
    "porteur_de_projet",
    "Porteur de projet",
    [
        Select2Field(
            "porteur",
            "Porteur de projet",
            choices=[],
            required=True,
            note="Il s'agit de préférence d'un personnel permanent de Sorbonne Université qui sera l’interlocuteur privilégié de Sorbonne Université et de la SATT LUTECH, le cas échéant.",
        )
    ],
)

exploitation = FieldSet(
    "exploitation",
    "Exploitation",
    [
        BooleanField(
            "negociations",
            "L’œuvre fait-elle l’objet ou a-t-elle fait l’objet de négociations pour son utilisation et/ou son exploitation (par un organisme de recherche ou par un industriel) ?",
            required=True,
        ),
        TextAreaField(
            "precision_negociations",
            "Précisez l’organisme de recherche ou l’industriel, le contexte, la nature du contrat",
        ),
    ],
)

engagement = FieldSet(
    "engagement",
    "Engagement de l'interlocuteur privilégié de l'invention",
    [
        Boolean2Field(
            "engagement",
            "En tant que porteur du projet, je reconnais accepter d’être le contact scientifique pour l’œuvre auprès de Sorbonne Université, et de la SATT LUTECH le cas échéant, au nom de l’ensemble des contributeurs et confirme la validité des renseignements indiqués dans cette déclaration de logiciel/base de données.",
            required=True,
        )
    ],
)



# from __future__ import annotations
#
# from typing import Set
#
# from labster.domain2.model.profile import Profile, ProfileId, ProfileRepository
#
# from .base import BaseInMemoryRepository
#
#
# class InmemoryProfileRepository(BaseInMemoryRepository, ProfileRepository):
#     def put(self, profile: Profile):
#         if not profile.id:
#             profile.id = ProfileId.new()
#         self._data[profile.id] = profile
#
#     def delete(self, profile: Profile):
#         del self._data[profile.id]
#
#     def get_all(self) -> Set[Profile]:
#         return set(self._data.values())
from __future__ import annotations



from __future__ import annotations

from labster.di import injector
from labster.domain2.services.roles import Role, RoleService
from labster.rpc.queries.demandes_tables import get_table_view
from labster.security import get_current_profile
from labster.types import JSON

SCOPES = [
    # DR&I
    "contact",
    "mes structures dri",
    "dri",
    "drv",
    # Autres structures
    "porteur",
    "gestionnaire",
    "mes structures",
]

SCOPES_RESPONSABLE = [
    # DR&I
    "contact",
    "mes structures dri",
    "dri",
    "drv",
    # Autres structures
    "mes structures",
    "porteur",
    "gestionnaire",
]


role_service = injector.get(RoleService)


def get_boxes(archives=False) -> JSON:
    archives = bool(archives)
    user = get_current_profile()

    if role_service.has_role(user, Role.RESPONSABLE, "*"):
        scopes = SCOPES_RESPONSABLE
    else:
        scopes = SCOPES
    result = []
    for scope in scopes:
        table_view = get_table_view(scope, user, archives)
        if table_view:
            result.append(
                {"title": table_view.title, "scope": scope, "archives": archives}
            )

    return result



from __future__ import annotations

import glob
import os
import sys

import structlog

from .ldif import parse_ldif_file, update_users_from_records
from .roles import RolesUpdater

logger = structlog.get_logger()


def sync_users():
    logger.info("## Synchronisation des utilisateurs avec le dump LDAP")
    ldif_file = get_ldif_file()
    new_records = parse_ldif_file(ldif_file)

    print(f"Nbre d'utilisateurs dans le LDIF: {len(new_records)}")
    update_users_from_records(new_records)

    logger.info("## Mise a jour des roles")
    update_roles()


def get_ldif_file():
    files = glob.glob("annuaire/export-Lab-Co-Dev*.ldif")
    if not files:
        files = glob.glob("annuaire/export-Lab-Co.*.ldif")
    files.sort()
    if not files:
        logger.error("Error: not LDIF file found in annuaire/")
        sys.exit(-1)
    ldif_file = files[-1]

    # Remove old files
    for file_to_remove in files[0:-1]:
        logger.info(f"removing {file_to_remove}")
        os.unlink(file_to_remove)

    return ldif_file


def update_roles(max=0):
    roles_updater = RolesUpdater()
    roles_updater.update_roles(max)



from __future__ import annotations

from flask_sqlalchemy import SQLAlchemy
from jsonrpcserver import method
from werkzeug.exceptions import BadRequest

from labster.di import injector
from labster.domain2.services.constants import save_constants

db = injector.get(SQLAlchemy)


@method
def update_constants(constants):
    if not constants:
        raise BadRequest()

    save_constants(constants)

    db.session.commit()



from __future__ import annotations

import os

import structlog
from flask import Flask, Request, flash, redirect, render_template, url_for
from flask_sqlalchemy import SQLAlchemy
from werkzeug.exceptions import Unauthorized

from labster.di import injector
from labster.domain2.model.profile import Profile
from labster.domain2.model.structure import StructureRepository
from labster.domain2.services.roles import Role, RoleService
from labster.security import get_current_user
from labster.util import sort_by_name

from . import route
from .cas import get_user_by_login, login_user

logger = structlog.get_logger()
db = injector.get(SQLAlchemy)
structure_repo = injector.get(StructureRepository)
role_service = injector.get(RoleService)


#
# Routes
#
@route("/backdoor")
def backdoor(app: Flask, request: Request):
    allow_backdoor = app.config.get("ALLOW_BACKDOOR") or os.getenv("ALLOW_BACKDOOR")
    if not (allow_backdoor or app.testing):
        raise Unauthorized()

    login = request.args.get("login", "poulainm")
    user = get_user_by_login(login)
    login_user(user)

    if "text/html" in request.headers.get("accept", ""):
        home_url = url_for(".login", _external=True)
        return redirect(home_url)
    else:
        return "", 201


@route("/switch", methods=["GET", "POST"])
def switch(app: Flask, request: Request):
    user = get_current_user()
    if user.is_anonymous:
        raise Unauthorized()

    profile = user.profile

    testing = app.config.get("TESTING", False)
    if not testing and not profile.has_role(Role.ADMIN_CENTRAL):
        raise Unauthorized()

    if request.args:
        return do_switch(request.args)

    gouvernance = get_users_by_login({"chambaz"})
    gestionnaires = get_users_by_login(
        {"pulcherie", "boyern", "courtoisi", "sos", "girardv"}
    )
    porteurs = get_users_by_login(
        {"carapezzi", "duhieu", "lombard", "diasdeamorim", "valdes"}
    )
    directeurs = get_users_by_login(
        {
            "santiardbaro",
            "charretier",
            "sciandra",
            "mercierc",
            "mouchelj",
            "stemmann",
        }
    )

    dri = structure_repo.get_by_sigle("DR&I")
    membres_dri = role_service.get_users_with_given_role(Role.MEMBRE_AFFECTE, dri)
    membres_dri = sort_by_name(membres_dri)
    membres_dri_logins = [m.login for m in membres_dri if m.active]
    membres_dri = get_users_by_login(membres_dri_logins)

    groups = [
        ["Gouvernance", gouvernance],
        ["Gestionnaires", gestionnaires],
        ["Porteurs", porteurs],
        ["Directeurs", directeurs],
        ["DR&I", membres_dri],
    ]
    return render_template("auth/login.j2", groups=groups)


def do_switch(args):
    login = args.get("login", None)

    new_user = get_user_by_login(login)

    if not new_user:
        flash(f"Utilisateur {login} inconnu.", "danger")
        return redirect(url_for(".switch", _external=True))

    login_user(new_user)
    return redirect(url_for(".login", _external=True))


# Utilisé en phase de test (URL: /switch/).
def get_users_by_login(logins) -> list[Profile]:
    users = (
        db.session.query(Profile)
        .filter(Profile.login.in_(logins))
        .filter(Profile.active == True)
        .order_by(Profile.login)
        .all()
    )

    for user in users:
        roles = role_service.get_roles_for_user(user)
        values = sorted(
            r.value for r in roles.keys() if not r.value.startswith("Membre ")
        )
        user._roles = ", ".join(values)

    return users



"""Formulaire Avenant Convention."""
from __future__ import annotations

from labster.newforms.base import HTML, BooleanField, DateField, FieldSet, \
    Form, IntegerField, ListePartenaires, Select2Field, StringField, \
    TextAreaField
#
# Fieldsets
#
from labster.newforms.common import contributeurs, laboratoire, \
    structures_concernees
from labster.newforms.convention import CHOICES2

porteur_de_projet = FieldSet(  #
    "porteur_de_projet",
    "Porteur de projet",
    [Select2Field("porteur", "Porteur de projet", choices=[], required=True)],
)

contrat = FieldSet(  #
    "contrat",
    "Contrat",
    [
        HTML("<p>Sur quel contrat porte l’avenant ?</p>"),
        StringField("nom_projet", "Nom ou acronyme du projet", required=True),
        StringField("eotp_ou_no_dgrtt", "eOTP ou № DR&I", required=True),
        Select2Field(
            "type_financeur", "Type de financeur", required=True, choices=CHOICES2
        ),
        BooleanField(
            "integre_entreprise",
            "Le projet intègre-t-il une entreprise&nbsp;?",
            required=True,
        ),
        #
        HTML(
            "<p>Votre demande concerne la modification... (plusieurs choix possibles)</p>"
        ),
        #
        BooleanField("montant", "Du montant"),
        IntegerField("nouveau_montant", "Indiquez les modifications du montant"),
        HTML(
            "<div class='row'><div class='col-md-offset-4 col-md-8'>Vous pouvez aussi modifier le calcul du coût complet du contrat avec l’onglet <b>Feuille de coût</b> ou joindre une nouvelle annexe financière avec l’onglet <b>Pièces à joindre</b></div></div>",
            name="message_montant",
        ),
        #
        BooleanField("duree", "De la durée"),
        DateField("nouvelle_date_fin", "Nouvelle date de fin"),
        #
        BooleanField("programme_scientifique", "Du programme scientifique"),
        HTML(
            "<div class='row'><div class='col-md-offset-4 col-md-8'>Utilisez l’onglet <b>Pièces à joindre</b> pour déposer le nouveau projet scientifique</div></div>",
            name="message_programme_scientifique",
        ),
        #
        BooleanField("consortium", "Du consortium"),
        BooleanField("retirer_partenaires", "Retirer un/des partenaire(s)"),
        TextAreaField("partenaires_a_retirer", "Partenaires à retirer"),
        BooleanField("ajouter_partenaires", "Ajouter un/des partenaire(s)"),
        ListePartenaires(
            "partenaires_a_ajouter", "Partenaires à ajouter", required=False
        ),
        BooleanField("modifier_partenaires", "Modifier un/des partenaire(s)"),
        ListePartenaires(
            "partenaires_a_modifier", "Partenaires à modifier", required=False
        ),
        #
        BooleanField("autre", "Autre"),
        TextAreaField("autre_precisez", "Précisez"),
    ],
)

commentaires = FieldSet(  #
    "commentaires",
    "Commentaires",
    [
        HTML(
            """
<p>Vous pouvez indiquer dans ce champs des éléments complémentaires nécessaires à l’instruction de votre demande.</p>
<p><em>Par exemple : besoin de recruter dès le démarrage du projet ; ce nouveau contrat s’inscrit-il dans un MTA et NDA existant, etc.</em></p>"""
        ),
        TextAreaField("commentaires", "Commentaires"),
    ],
)


#
# The Form
#
class DemandeAvenantConventionForm(Form):
    template = "newforms/newform.html"
    name = "avenant_convention"
    label = "Avenant à une convention de recherche"

    fieldsets = [
        laboratoire,
        porteur_de_projet,
        structures_concernees,
        contributeurs,
        #
        contrat,
        commentaires,
    ]



from __future__ import annotations

from datetime import date

from labster.domain2.model.util import parse_date


def test_parsedate() -> None:
    assert parse_date("2014/01/10") == date(2014, 1, 10)
    assert parse_date("2014-01-10") == date(2014, 1, 10)
    assert parse_date("10-01-2014") == date(2014, 1, 10)
    assert parse_date("10/01/2014") == date(2014, 1, 10)



"""Formulaire RH."""
from __future__ import annotations

import re
from copy import deepcopy

from flask_sqlalchemy import SQLAlchemy

from labster.di import injector
from labster.domain2.model.structure import Structure
from labster.domain.services.constants import get_constant
from labster.newforms.base import BooleanField, DateField, EmailField, \
    FieldSet, Form, IntegerField, Select2Field, StringField, TextAreaField
from labster.newforms.common import laboratoire, structures_concernees

# acceptation_principe_recrutement = BooleanField(
#     "J'ai pris connaissance et j'accepte les principes de recrutement de Sorbonne Université",
#     validators=[required()])

db = injector.get(SQLAlchemy)

CHOICES1 = ["CDD", "Doctorant", "Bourse Marie Curie"]

CHOICES2 = [
    "Contrat initial",
    "Renouvellement",
    "Modification du contrat en cours",
    "Contrat doctoral initial",
    "Prolongation par avenant d'un contrat doctoral en cours",
    "Prolongation par CDD d'un doctorat en cours (cas particulier)",
    "Thèse medico-scientifique",
]

nature = FieldSet(
    "nature",
    "Nature et type de la demande",
    [
        Select2Field(
            "nature_du_recrutement",
            "Nature du recrutement",
            choices=CHOICES1,
            required=True,
        ),
        Select2Field(
            "type_de_demande",
            "Type de demande",
            choices=CHOICES2,
            show_if="false",
            required=True,
        ),
    ],
)


def choice6():
    ecoles_doctorales = (
        db.session.query(Structure).filter_by(type_name="École doctorale").all()
    )

    def sorter(ed):
        m = re.search("([0-9]+)", ed.nom)
        if m:
            n = int(m.group(1))
        else:
            n = 0
        return n

    ecoles_doctorales.sort(key=sorter)
    return [""] + [ed.nom for ed in ecoles_doctorales]


responsable_scientifique = FieldSet(  #
    "responsable_scientifique",
    "Responsable scientifique de la personne recrutée",
    [
        # TODO
        Select2Field(
            "ecole_doctorale", "École doctorale", choices=choice6, required=True
        ),
        Select2Field("porteur", "Nom", choices=[], required=True),
    ],
)

CHOICES3 = ["eOTP", "№ DR&I", "Notification de financement"]

CHOICES4 = ["eOTP", "№ DR&I", "1/2 allocation IFD", "Notification de financement"]

contrat = FieldSet(  #
    "contrat",
    "Contrat de recherche finançant cette demande de recrutement",
    [
        Select2Field("financement", "Financement", choices=CHOICES3, required=True),
        StringField("numero_de_financement", "Numéro du financement", required=True),
        BooleanField("co_finance", "Ce recrutement est-il co-financé&nbsp;?"),
        Select2Field("financement2", "Financement 2", choices=CHOICES4, required=True),
        StringField("numero_de_financement2", "Numéro du financement"),
        Select2Field("structure_financeuse", "Structure", choices=[]),
    ],
)

CHOICES5 = ["M.", "Mme"]

# beware, the exact strings are used in rh.js.
CHOICES_SIMILAR_EXP = [
    "Pour un post-doctorant: doctorat",
    "Pour un post-doctorant: doctorat plus 2 ans",
    "Pour un post-doctorant: doctorat plus 3 ans",
    "Pour un ingénieur, technicien ou administratif",
    "Pour un chercheur",
]

candidat = FieldSet(  #
    "candidat",
    "Candidat",
    [
        Select2Field("civilite", "Civilité", choices=CHOICES5, required=True),
        StringField("nom", "Nom", required=True),
        StringField("prenom", "Prénom", required=True),
        EmailField("email", "Mail", required=True),
        Select2Field(
            "similar_experience",
            "Expérience professionnelle",
            choices=CHOICES_SIMILAR_EXP,
            required=True,
            note="Des documents sont à joindre à votre demande. "
            "Une fois que vous l'aurez enregistrée, utilisez l'onglet "
            "<b>Pièces à joindre<b> pour consulter la liste et envoyer les "
            "documents.",
        ),
        IntegerField(
            "similar_experience_years",
            "Indiquez le nombre d'années d'expérience sur des niveaux de "
            "mission comparables",
        ),
    ],
)


def choice7():
    return get_constant("recrutement.grades")


CHOICES8 = [
    "Paris ou région parisienne",
    "Banyuls-sur-mer",
    "Roscoff",
    "Villefranche-sur-mer",
]

CHOICES9 = ["100%", "90%", "80%", "70%", "60%", "50%"]

poste = FieldSet(  #
    "poste",
    "Poste à pourvoir",
    [
        BooleanField(
            "modification_mission",
            "La modification concerne-t-elle les missions du contrat initial ?",
        ),
        StringField("fonction_du_poste", "Fonction du poste", required=True),
        Select2Field(
            "grade_correspondant",
            "Équivalent corps/grade fonction publique",
            choices=choice7,
        ),
        StringField(
            "objet_de_la_mission", "Objet de la mission / de l'étude", required=True
        ),
        Select2Field("localisation", "Localisation", choices=CHOICES8, required=True),
        Select2Field(
            "quotite_de_travail", "Quotité de travail", choices=CHOICES9, required=True
        ),
    ],
)

dates = FieldSet(  #
    "dates",
    "Dates du contrat",
    [
        DateField("date_debut", "Date de début", required=True),
        DateField("date_fin", "Date de fin", required=True),
    ],
)

autre_modification = FieldSet(  #
    "autre_modification",
    "Autre modification",
    [
        BooleanField(
            "modification_autre",
            "L’avenant modifie-t-il un autre élément du contrat initial&nbsp;?",
        ),
        TextAreaField("modification_autre_detail", "Indiquez lequel"),
    ],
)

publicite = FieldSet(  #
    "publicite",
    "Publicité",
    [
        TextAreaField(
            "publicite",
            "Comment avez-vous fait la publicité de ce poste (publication sur le "
            "site de la structure, le site de Sorbonne Université, ...)&nbsp;?",
            required=True,
        ),
        IntegerField(
            "nb_candidats_recus",
            "Combien de candidats avez-vous reçu en entretien&nbsp;?",
            required=True,
        ),
    ],
)

commentaire = FieldSet(  #
    "commentaire", "Commentaire", [TextAreaField("commentaire", "Commentaire")]
)


#
# The Form
#
class DemandeRHForm(Form):
    name = "rh"
    template = "newforms/newform.html"
    label = "Recrutement"

    def post_init(self):
        remuneration = FieldSet(  #
            "remuneration",
            "Rémunération",
            [
                BooleanField(
                    "modification_remuneration",
                    "L’avenant modifie-t-il les conditions de rémunération&nbsp;?",
                ),
                IntegerField(
                    "salaire_brut_mensuel_indicatif",
                    "Montant indicatif du salaire brut mensuel (en Euros)",
                    note=get_constant("demande_recrutement.bareme"),
                    editable=False,
                ),
                IntegerField(
                    "salaire_brut_mensuel",
                    "Montant du salaire brut mensuel (en Euros)",
                    required=True,
                ),
                TextAreaField("salaire_justification", "Justifier le salaire"),
                IntegerField(
                    "salaire_brut_mensuel_etp",
                    "Pour information, montant du brut mensuel équivalent "
                    "à un temps plein (en Euros)",
                    editable=False,
                ),
                TextAreaField("justification_du_salaire", "Justification du salaire"),
                BooleanField(
                    "indemnite_transport_en_commun", "Indemnité de transport en commun"
                ),
                IntegerField("nombre_enfants_a_charge", "Nombre d'enfants à charge"),
            ],
        )

        fieldsets = [
            nature,
            laboratoire,
            responsable_scientifique,
            structures_concernees,
            contrat,
            candidat,
            poste,
            dates,
            remuneration,
            autre_modification,
            publicite,
            commentaire,
        ]

        self.fieldsets = deepcopy(fieldsets)

        conditions = get_constant("recrutement.principes")
        self.conditions = conditions



from __future__ import annotations

from typing import TYPE_CHECKING

from flask import g

from labster.domain2.model.profile import Profile as NewProfile

if TYPE_CHECKING:
    from labster.security import User


class AuthContext:
    @property
    def current_user(self) -> User:
        try:
            return g.current_user
        except (RuntimeError, AttributeError):
            from labster.security import AnonymousUser

            return AnonymousUser()

    @property
    def current_profile(self) -> NewProfile | None:
        try:
            return g.current_profile
        except (RuntimeError, AttributeError):
            return None



from __future__ import annotations

from flask import render_template
from flask_mail import Message

from labster.extensions import mail


def send_notification(user, body: str, workflow):
    """
    :param user: user to notify
    :param body: message body (HTML)
    :param workflow:

    :return: the notification object
    """
    raise RuntimeError

    # from labster.domain2.model.notification import Notification
    #
    # notification = Notification(
    #     user=user, body=body, demande=workflow.case, actor=workflow.actor
    # )
    # try:
    #     app = db.get_app()
    # except RuntimeError:
    #     app = None
    # if app:
    #     db.session.add(notification)
    #
    # if user.preferences_notifications in (FLUX_TENDU, None):
    #     notification.sent = True
    #     send_notification_by_email(notification)
    #
    # return notification


# def send_notification_by_email(notification) -> None:
#     user = notification.user
#     subject = "Une action a été réalisée sur une de vos demandes Lab&Co"
#     recipients = [user.email]
#     ctx = {
#         "notification": notification,
#         "demande": notification.demande,
#         "now": datetime.now(),
#     }
#     html = render_template("emails/notif-generique.html", **ctx)
#     msg = Message(subject, recipients=recipients, html=html)
#     mail.send(msg)


def send_email(recipients, subject, template, context):
    if not isinstance(recipients, list):
        recipients = [recipients]

    recipients = [r.email for r in recipients]
    html = render_template("emails/" + template, **context)
    msg = Message(subject, recipients=recipients, html=html)
    mail.send(msg)



from __future__ import annotations

import json
import sys
import time
import traceback
from json import JSONDecodeError
from pprint import pformat, pprint

import sentry_sdk
from flask import Blueprint, Flask
from flask import Request as FlaskRequest
from flask import Response
from jsonrpcserver.dispatcher import call, create_requests, schema
from jsonrpcserver.methods import global_methods
from jsonrpcserver.request import Request
from jsonrpcserver.response import ErrorResponse, ExceptionResponse, \
    InvalidJSONResponse, InvalidJSONRPCResponse
from jsonrpcserver.response import Response as JsonRpcResponse
from jsonrpcserver.response import SuccessResponse
from jsonschema import ValidationError
from jsonschema.validators import validator_for
from werkzeug.exceptions import Forbidden, Unauthorized

from labster.auth import AuthContext
from labster.types import JSON

from ..domain2.services.roles import Role
from .registry import register_submodules

blueprint = Blueprint("rpc", __name__, url_prefix="/rpc")
route = blueprint.route

validator_class = validator_for(schema)
validator_class.check_schema(schema)
validator = validator_class(schema)

timer_d = {}


@blueprint.record
def configure(state):
    register_submodules()


@route("/", methods=["POST"])
def index(app: Flask, request: FlaskRequest, auth: AuthContext):
    req = request.get_data(as_text=True)
    req_json = json.loads(req)
    if app.debug:
        pprint(req_json)

    user = auth.current_user
    if user and not user.is_authenticated:
        raise Unauthorized()

    return run(req, app)


@route("/debug/<method_name>")
def debug(
    method_name: str, app: Flask, request: FlaskRequest, auth_context: AuthContext
):
    user = auth_context.current_profile
    if user and not user.has_role(Role.ADMIN_CENTRAL) and not app.debug:
        raise Forbidden()

    args = dict(request.args)
    req_dict = {"id": 0, "jsonrpc": "2.0", "method": method_name, "params": args}
    req = json.dumps(req_dict)
    return run(req, app)


def run(req: str, app: Flask) -> Response:
    if app.debug:
        print(78 * "#")
        print("# RPC request:")
        req_json = json.loads(req)
        print(78 * "#")
        sys.stdout.flush()
        timer_d[req_json["id"]] = time.time()

        response = dispatch(req)

        print(78 * "#")
        print("# RPC response:")

        print("request:")
        s = pformat(req_json)
        if len(s) < 200:
            print(s)
        else:
            print(s[0:200] + "...")

        print("response:")
        s = pformat(response.deserialized())
        if isinstance(response, ErrorResponse):
            print(s)
        elif len(s) < 200:
            print(s)
        else:
            print(s[0:200] + "...")
        if hasattr(response, "exc"):
            print(response.exc)

        req_id = req_json["id"]
        if req_id in timer_d:
            dt = 1000 * (time.time() - timer_d[req_id])
            print(f"Elapsed time: {dt:.2f}ms")
            del timer_d[req_id]

        print(78 * "#")
        sys.stdout.flush()

    else:
        response = dispatch(req)

    return Response(str(response), response.http_status, mimetype="application/json")


def dispatch(request_raw: str) -> JsonRpcResponse:
    """Dispatch a request (or requests) to methods.

    This is the main public method, it's the only one with optional params, and the only
    one that can be configured with a config file/env vars.

    Args:
        request_raw: The incoming request string.

    Returns:
        A Response.
    """
    methods = global_methods
    try:
        request_json: JSON = json.loads(request_raw)
        validator.validate(request_json)
    except JSONDecodeError as exc:
        return InvalidJSONResponse(data=str(exc), debug=True)
    except ValidationError:
        return InvalidJSONRPCResponse(data=None)

    request = create_requests(request_json, convert_camel_case=False)
    assert isinstance(request, Request)

    try:
        method_name = request.method
        method = methods.items[method_name]

        with sentry_sdk.start_span(op="rpc", transaction="rpc." + method_name) as span:
            span.set_data("args", request.args)
            span.set_data("kwargs", request.kwargs)
            result = call(method, *request.args, **request.kwargs)

        return SuccessResponse(result=result, id=request.id)

    except Exception as exc:
        traceback.print_exc()
        sys.stdout.flush()
        sys.stderr.flush()
        return ExceptionResponse(exc, id=request.id, debug=True)



from __future__ import annotations

from jsonrpcserver import method

from labster.domain2.services.constants import TYPES
from labster.domain2.services.constants import get_constants as _get_constants
from labster.types import JSONDict


@method
def get_constants() -> JSONDict:
    return {"values": _get_constants(), "types": TYPES}



from __future__ import annotations

from abc import ABC, ABCMeta
from datetime import datetime
from typing import TYPE_CHECKING, Any
from uuid import uuid4

from abilian.app import db
from sqlalchemy import JSON, Boolean, Column, DateTime, Integer, String

from labster.domain2.model.base import Repository
from labster.ldap.constants import DRI_DN

if TYPE_CHECKING:
    from labster.domain2.model.structure import Structure
    from labster.domain2.services.roles import Role


FLUX_TENDU = 0
DAILY = 1
WEEKLY = 2


class ProfileId(str):
    @staticmethod
    def new() -> ProfileId:
        return ProfileId(uuid4())


# @attrs(eq=False, order=False, repr=False, auto_attribs=True)
class Profile(db.Model):
    __tablename__ = "v3_profiles"

    id = Column(String(36), primary_key=True)

    uid = Column(String(64), unique=True, nullable=True)
    old_id = Column(Integer, unique=True, nullable=True)
    old_uid = Column(String(64), unique=True, nullable=True)
    login = Column(String(64), default="", nullable=False)

    nom = Column(String, default="", nullable=False)
    prenom = Column(String, default="", nullable=False)
    email = Column(String, default="", nullable=False)
    adresse = Column(String, default="", nullable=False)
    telephone = Column(String, default="", nullable=False)

    active = Column(Boolean, default=False, nullable=False)
    affectation = Column(String, default="", nullable=False)
    fonctions = Column(JSON, nullable=False)

    preferences_notifications = Column(Integer, default=0, nullable=False)
    preferences_nb_jours_notifications = Column(Integer, default=0)

    date_derniere_notification_vue = Column(DateTime, default=datetime.utcnow)

    def __init__(self, **kw):
        self.id = str(uuid4())
        self.nom = ""
        self.prenom = ""
        self.fonctions = []
        super().__init__(**kw)

    def __str__(self):
        return f"<Profile {self.full_name}>"

    def __repr__(self):
        return str(self)

    @property
    def full_name(self):
        return self.prenom + " " + self.nom

    @property
    def reversed_name(self):
        return f"{self.nom}, {self.prenom}"

    @property
    def name(self):
        return self.full_name

    def activate(self):
        self.active = True

    def deactivate(self):
        self.active = False

    #
    # Roles
    #
    def structure_d_appartenance(self) -> Structure:
        from labster.di import injector
        from labster.domain2.services.roles import Role, RoleService

        role_service = injector.get(RoleService)
        roles_dict = role_service.get_roles_for_user(self)
        structures = roles_dict[Role.MEMBRE_AFFECTE]
        assert len(structures) == 1
        return list(structures)[0]

    def has_role(self, role: Role, context: Any = None) -> bool:
        from labster.di import injector
        from labster.domain2.services.roles import RoleService

        role_service = injector.get(RoleService)
        return role_service.has_role(self, role, context)

    def is_membre_dri(self):
        from labster.di import injector
        from labster.domain2.model.structure import StructureRepository
        from labster.domain2.services.roles import Role

        structure_repo = injector.get(StructureRepository)
        dri = structure_repo.get_by_dn(DRI_DN)
        return self.has_role(Role.MEMBRE, dri)


class ProfileRepository(Repository, ABC, metaclass=ABCMeta):
    def get_all(self) -> set[Profile]:
        raise NotImplementedError

    def put(self, profile: Profile):
        raise NotImplementedError

    def delete(self, profile: Profile):
        raise NotImplementedError

    def get_by_id(self, id: ProfileId) -> Profile:
        raise NotImplementedError

    def get_by_uid(self, uid: str) -> Profile:
        raise NotImplementedError

    def get_by_old_id(self, old_id: int) -> Profile:
        raise NotImplementedError

    def get_by_login(self, login: str) -> Profile:
        raise NotImplementedError

    def get_by_old_uid(self, old_uid: str) -> Profile:
        raise NotImplementedError



from __future__ import annotations


class Repository:
    def put(self, obj) -> None:
        raise NotImplementedError

    def clear(self) -> None:
        raise NotImplementedError

    def is_empty(self) -> bool:
        raise NotImplementedError



from __future__ import annotations

from jsonrpcserver import method

from labster.domain2.model.demande import DemandeAutre, \
    DemandeAvenantConvention, DemandeConvention, DemandePiMixin, DemandeRH
from labster.domain2.services.roles import Role
from labster.domain2.services.workflow.states import EN_VALIDATION
from labster.rpc.queries.demandes_tables import mes_taches
from labster.security import get_current_profile


@method
def get_nb_demandes_a_valider():
    user = get_current_profile()

    ctx = {
        "total_demandes": 0,
        "nb_pi_a_valider": 0,
        "nb_recrutements_a_valider": 0,
        "nb_conventions_a_valider": 0,
        "nb_autres_a_valider": 0,
    }

    if user.has_role(Role.RESPONSABLE, "*"):
        demandes = mes_taches(user)
        demandes = [d for d in demandes if d.wf_state == EN_VALIDATION.id]

        conventions = [
            d
            for d in demandes
            if isinstance(d, (DemandeConvention, DemandeAvenantConvention))
        ]
        recrutements = [d for d in demandes if isinstance(d, DemandeRH)]
        pi = [d for d in demandes if isinstance(d, DemandePiMixin)]
        autres = [d for d in demandes if isinstance(d, DemandeAutre)]

        ctx["total_demandes"] = len(demandes)
        ctx["nb_conventions_a_valider"] = len(conventions)
        ctx["nb_recrutements_a_valider"] = len(recrutements)
        ctx["nb_pi_a_valider"] = len(pi)
        ctx["nb_autres_a_valider"] = len(autres)

    return ctx



# TODO: this is still being worked out by the mypy team
from __future__ import annotations

from typing import Any, Sequence, Union

JSON = Union[str, int, float, bool, None, dict[str, Any], list[Any]]
JSONDict = dict[str, JSON]
JSONList = Sequence[JSON]



from __future__ import annotations

from flask_sqlalchemy import SQLAlchemy
from injector import inject

# types_demande = [cls._type.value for cls in _REGISTRY.values()]


class Mapper:
    @inject
    def __init__(self, db: SQLAlchemy):
        self.db = db
        self.metadata = db.metadata

        # self.map_profiles()
        # self.map_structures()
        # self.map_demandes()

    # def map_profiles(self):
    #     profiles = Table(
    #         "v3_profiles",
    #         self.db.metadata,
    #         #
    #         Column("id", String(36), primary_key=True),
    #         Column("uid", String(64), unique=True, nullable=True),
    #         Column("old_id", Integer, unique=True, nullable=True),
    #         Column("old_uid", String(64), unique=True, nullable=True),
    #         Column("login", String(64), default="", nullable=False),
    #         #
    #         Column("nom", String, default="", nullable=False),
    #         Column("prenom", String, default="", nullable=False),
    #         Column("email", String, default="", nullable=False),
    #         Column("adresse", String, default="", nullable=False),
    #         Column("telephone", String, default="", nullable=False),
    #         #
    #         Column("active", Boolean, default=False, nullable=False),
    #         Column("affectation", String, default="", nullable=False),
    #         Column("fonctions", JSON, nullable=False),
    #         #
    #         Column("preferences_notifications", Integer, default=0, nullable=False),
    #     )
    #
    #     mapper(Profile, profiles)
    #
    # def map_structures(self):
    #     hierarchy = Table(
    #         "v3_hierarchy",
    #         self.db.metadata,
    #         Column("parent_id", String(36), ForeignKey("v3_structures.id")),
    #         Column("child_id", String(36), ForeignKey("v3_structures.id")),
    #     )
    #
    #     structures = Table(
    #         "v3_structures",
    #         self.db.metadata,
    #         #
    #         Column("id", String(36), primary_key=True),
    #         Column("old_id", Integer),
    #         Column("active", Boolean),
    #         Column("type_name", String),
    #         #
    #         Column("nom", String),
    #         Column("sigle", String),
    #         Column("dn", String),
    #         Column("old_dn", String),
    #         Column("email", String),
    #         #
    #         Column("permettre_reponse_directe", Boolean),
    #         Column("permettre_soummission_directe", Boolean),
    #         #
    #     )
    #     mapper(
    #         Structure,
    #         structures,
    #         properties={
    #             "children": relationship(
    #                 Structure,
    #                 secondary=hierarchy,
    #                 primaryjoin=(hierarchy.c.parent_id == structures.c.id),
    #                 secondaryjoin=(hierarchy.c.child_id == structures.c.id),
    #                 collection_class=set,
    #                 backref=backref("parents", collection_class=set),
    #             )
    #         },
    #     )
    #
    # def map_demandes(self):
    #     demandes = Table(
    #         "v3_demandes",
    #         self.db.metadata,
    #         #
    #         Column("id", Integer, primary_key=True),
    #         Column("old_id", Integer),
    #         Column(
    #             "type",
    #             Enum(*types_demande, name="type_demande"),
    #             nullable=False,
    #             index=True,
    #         ),
    #         Column("created_at", DateTime),
    #         Column("updated_at", DateTime),
    #         #
    #         Column("nom", String),
    #         Column("name", String),
    #         Column("active", Boolean),
    #         Column("editable", Boolean),
    #         Column("no_infolab", String),
    #         Column("no_eotp", String),
    #         # Relations
    #         Column("contact_labco_id", String(36), ForeignKey("v3_profiles.id")),
    #         Column("gestionnaire_id", String(36), ForeignKey("v3_profiles.id")),
    #         Column("porteur_id", String(36), ForeignKey("v3_profiles.id")),
    #         Column("structure_id", String(36), ForeignKey("v3_structures.id")),
    #         #
    #         Column("data", JSON),
    #         Column("past_versions", JSON),
    #         Column("form_state", JSON),
    #         Column("attachments", JSON),
    #         Column("feuille_cout", JSON),
    #         Column("data", JSON),
    #         Column("documents_generes", JSON),
    #         # Workflow
    #         # wf_state = Column(WF_ENUM, default=EN_EDITION.id, nullable=False, index=True)
    #         Column("wf_state", String),
    #         Column("wf_date_derniere_action", DateTime),
    #         Column("wf_retard", Integer),
    #         Column("wf_history", JSON),
    #         Column("wf_data", JSON),
    #         Column("wf_state", String),
    #         # wf_stage_id = Column(Integer, ForeignKey(OrgUnit.id), index=True, nullable=True)
    #         # wf_stage = relationship(
    #         #     OrgUnit, primaryjoin=remote(Entity.id) == foreign(wf_stage_id)
    #         # )
    #         #
    #         # #: id de la personne responsable de la tâche en cours
    #         # wf_current_owner_id = Column(
    #         #     Integer, ForeignKey(Profile.id), index=True, nullable=True
    #         # )
    #         # #: la personne responsable de la tâche en cours
    #         # wf_current_owner = relationship(
    #         #     Profile, primaryjoin=remote(Entity.id) == foreign(wf_current_owner_id)
    #         # )
    #     )
    #
    #     properties = {
    #         "contact_labco": relationship(
    #             Profile, foreign_keys=[demandes.c.contact_labco_id]
    #         ),
    #         "porteur": relationship(Profile, foreign_keys=[demandes.c.porteur_id]),
    #         "gestionnaire": relationship(
    #             Profile, foreign_keys=[demandes.c.gestionnaire_id]
    #         ),
    #         "structure": relationship(Structure),
    #     }
    #     mapper(
    #         Demande,
    #         demandes,
    #         polymorphic_on=demandes.c.type,
    #         polymorphic_identity="none",
    #         properties=properties,
    #     )
    #     for cls in _REGISTRY.values():
    #         mapper(
    #             cls, demandes, inherits=Demande, polymorphic_identity=cls._type.value,
    #         )



from __future__ import annotations

from sqlalchemy import Column, Date, ForeignKey, Integer, String, Unicode

from labster.domain2.model.profile import Profile
from labster.domain2.model.structure import Structure
from labster.extensions import db


class StatsLine(db.Model):
    __tablename__ = "v3_stats_line"

    demande_id = Column(Integer, primary_key=True)

    # Filtrage
    type_demande = Column(String, index=True, nullable=False)

    date_soumission = Column(Date, index=True)
    date_finalisation = Column(Date, index=True)

    # structure_id = Column(Integer, index=True)
    # porteur_id = Column(Integer, index=True)
    porteur_id = Column(String(36), ForeignKey(Profile.id), index=True)
    structure_id = Column(String(36), ForeignKey(Structure.id), index=True)

    # Champs spécifiques
    financeur = Column(Unicode, index=True)
    type_recrutement = Column(Unicode, index=True)

    wf_state = Column(Unicode, index=True)

    # Drilldown
    l1 = Column(String(36), ForeignKey(Structure.id), index=True)
    l2 = Column(String(36), ForeignKey(Structure.id), index=True)
    l3 = Column(String(36), ForeignKey(Structure.id), index=True)
    l4 = Column(String(36), ForeignKey(Structure.id), index=True)
    l5 = Column(String(36), ForeignKey(Structure.id), index=True)
    l6 = Column(String(36), ForeignKey(Structure.id), index=True)
    l7 = Column(String(36), ForeignKey(Structure.id), index=True)
    l8 = Column(String(36), ForeignKey(Structure.id), index=True)
    l9 = Column(String(36), ForeignKey(Structure.id), index=True)
    l10 = Column(String(36), ForeignKey(Structure.id), index=True)

    # Agregations / Demande Conventions
    montant = Column(Integer)
    recrutements_prev = Column(Integer)

    # Agregations / Demandes RH
    salaire_brut_mensuel = Column(Integer)
    cout_total_mensuel = Column(Integer)

    # Commun Conventions / RH
    duree = Column(Integer)

    # Autres
    duree_traitement = Column(Integer)



from __future__ import annotations

from jsonrpcserver import method

from labster.types import JSONDict

from .registry import registry


@method
def get_context(name: str, params: JSONDict | None = None, **kw) -> JSONDict:
    if params is None:
        params = {}

    if kw:
        params = kw

    if name in registry:
        return registry[name](**params)
    else:
        return {}



from __future__ import annotations

import logging
import sqlite3
import time
import uuid
from pathlib import Path

import structlog
from flask import Flask, g, request
from structlog.dev import set_exc_info
from structlog.processors import JSONRenderer, StackInfoRenderer, \
    TimeStamper, format_exc_info


def init_logging(app: Flask) -> None:

    log_db = app.config.get("LOG_DB")
    if not log_db:
        return

    log_db_path = str(Path(app.instance_path) / log_db)
    print(log_db_path)

    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    logger.handlers[0].setLevel(logging.INFO)
    sh = SQLiteHandler(log_db_path)
    # sh.setLevel(logging.INFO)
    logger.addHandler(sh)

    processors = [
        StackInfoRenderer(),
        set_exc_info,
        format_exc_info,
        TimeStamper(fmt="%Y-%m-%d %H:%M.%S", utc=False),
        JSONRenderer(),
    ]

    structlog.configure(
        processors=processors,
        context_class=structlog.threadlocal.wrap_dict(dict),
        logger_factory=structlog.stdlib.LoggerFactory(),
    )

    @app.before_request
    def inject_request_id():
        request_id = str(uuid.uuid4())
        logger = structlog.get_logger(request_id=request_id, path=request.path)
        g.request_id = request_id

        logger.debug("Request started.")

    @app.after_request
    def log_end_of_request(response):
        logger = structlog.get_logger()
        logger.debug("Request ended successfully.")
        return response


class MyRenderer:
    key_order = ["request_id", "path", "user"]

    def __call__(self, _, __, event_dict):
        event = event_dict.pop("event")
        result = event
        items = self.ordered_items(event_dict)
        if items:
            result += " | " + " ".join(k + "=" + repr(v) for k, v in items)
        return result

    def ordered_items(self, event_dict):
        items = []
        for key in self.key_order:
            value = event_dict.pop(key, None)
            if value is not None:
                items.append((key, value))
        items += event_dict.items()
        return items


# class LogEntry(db.Model):
#     date = db.DateTime(nullable=False)
#     request_id = db.String(nullable=False)
#     _data = db.JSON(nullable=False)
#
#
# class DbLogger:
#     def msg(self, message):
#         entry = LogEntry()
#         print(message)
#
#
# class DbLoggerFactory:
#     def __call__(self, *args):
#         return DbLogger()

initial_sql = """CREATE TABLE IF NOT EXISTS log(
                    TimeStamp TEXT,
                    Source TEXT,
                    LogLevel INT,
                    LogLevelName TEXT,
                    Message TEXT,
                    Args TEXT,
                    Module TEXT,
                    FuncName TEXT,
                    LineNo INT,
                    Exception TEXT,
                    Process INT,
                    Thread TEXT,
                    ThreadName TEXT
               )"""

insertion_sql = """INSERT INTO log(
                    TimeStamp,
                    Source,
                    LogLevel,
                    LogLevelName,
                    Message,
                    Args,
                    Module,
                    FuncName,
                    LineNo,
                    Exception,
                    Process,
                    Thread,
                    ThreadName
               )
               VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
               """


class SQLiteHandler(logging.Handler):
    """Thread-safe logging handler for SQLite."""

    def __init__(self, db_file):
        logging.Handler.__init__(self)
        self.formatter = logging.Formatter()
        self.db_file = db_file
        conn = sqlite3.connect(db_file)
        conn.execute(initial_sql)
        conn.close()

    def format_time(self, record):
        """Create a time stamp."""
        record.dbtime = time.strftime(
            "%Y-%m-%d %H:%M:%S", time.localtime(record.created)
        )

    def emit(self, record):
        # self.format(record)
        self.format_time(record)
        if record.exc_info:  # for exceptions
            record.exc_text = self.formatter.formatException(record.exc_info)
        else:
            record.exc_text = ""

        # Insert the log record
        r = record
        value = (
            r.dbtime,
            r.name,
            r.levelno,
            r.levelname,
            r.msg,
            str(r.args),
            r.module,
            r.funcName,
            r.lineno,
            r.exc_text,
            r.process,
            r.thread,
            r.threadName,
        )
        conn = sqlite3.connect(self.db_file)
        conn.execute(insertion_sql, value)
        conn.close()



"""Main blueprint."""
from __future__ import annotations

import mimetypes
from datetime import datetime

import magic
from abilian.core.models.blob import Blob
from flask import Blueprint, Request, make_response, redirect, \
    render_template, request
from flask_sqlalchemy import SQLAlchemy
from werkzeug.exceptions import Forbidden, NotFound

from labster.domain2.model.demande import Demande, DemandeRH
from labster.domain2.model.profile import Profile
from labster.domain2.model.structure import StructureRepository
from labster.domain2.services.constants import get_constants
from labster.domain2.services.documents_generes import devis_rh, \
    lettre_commande_rh
from labster.rbac import acces_restreint, check_can_add_pj, \
    check_read_access, feuille_cout_editable
from labster.rpc.commands.demandes import cleanup_model
from labster.security import get_current_profile, login_required

blueprint = Blueprint("main", __name__, url_prefix="")
route = blueprint.route

__all__ = ()


@blueprint.before_request
@login_required
def before_request():
    pass


@route("/")
def home() -> str:
    return render_template("v3.j2")


@route("/calculettes/feuille_cout")
def calculette_feuille_cout():
    constants = get_constants()
    model = {
        "id": None,
        "editable": True,
        "constants": constants["convention"],
        "couts_charges": constants["recrutement"]["couts_charges"],
    }
    model["constants"]["POINT_INDICE"] = constants["point_indice"]

    return render_template("calculette_feuille_cout.html", model=model)


@route("/feuille_cout/<id>")
def feuille_cout(id: int, db: SQLAlchemy):
    constants = get_constants()

    demande = db.session.query(Demande).get(id)
    check_read_access(demande)

    model = demande.feuille_cout or {}
    model["id"] = demande.id
    model["editable"] = feuille_cout_editable(demande)
    model["constants"] = constants["convention"]
    model["constants"]["POINT_INDICE"] = constants["point_indice"]
    model["couts_charges"] = constants["recrutement"]["couts_charges"]

    return render_template("calculette_feuille_cout.html", model=model)


@route("/blob/<int:demande_id>/<int:blob_id>")
def blob(demande_id: int, blob_id: int, db: SQLAlchemy):
    demande = db.session.query(Demande).get(demande_id)
    blob = Blob.query.get(blob_id)
    if not demande or not blob:
        raise NotFound()

    check_read_access(demande)

    def get_filename(demande, blob_id):
        for filename, v in demande.attachments.items():
            if v["id"] == blob_id:
                return filename

        for d in demande.documents_generes:
            if d["blob_id"] == blob_id:
                return d["name"]

        return "fichier inconnu"

    filename = get_filename(demande, blob_id)
    suffix = get_suffix_for(blob.value or b"")
    if not suffix:
        suffix = ".pdf"
    if not filename.endswith(suffix):
        filename = filename + suffix

    response = make_response(blob.value or "")
    if blob.value:
        response.headers["content-type"] = magic.from_buffer(blob.value, mime=True)
    else:
        response.headers["content-type"] = "text/plain"
    content_disposition = f'attachment;filename="{filename}"'
    response.headers["content-disposition"] = content_disposition.encode()
    return response


def get_suffix_for(value: bytes):
    mime_type = magic.from_buffer(value, mime=True)
    extension = mimetypes.guess_extension(mime_type)
    return extension


@route("/upload/", methods=["POST"])
def upload(request: Request, db: SQLAlchemy):
    user = get_current_profile()
    form = request.form
    demande_id = form["demande_id"]
    demande = db.session.query(Demande).get(demande_id)
    check_can_add_pj(demande)

    files = request.files
    for file in files.values():
        file_name = file.filename

        data = file.read()
        blob = Blob(data)
        db.session.add(blob)
        db.session.flush()

        demande.attachments[file_name] = {
            "date": datetime.now().isoformat(),
            "id": blob.id,
            "creator": user.login,
        }

    db.session.commit()
    return "OK"


@route("/demandes/<int:id>/devis_rh")
def devis_rh_rest(id, db: SQLAlchemy):
    demande = db.session.query(Demande).get(id)
    if acces_restreint(demande):
        raise Forbidden()

    assert isinstance(demande, DemandeRH)

    response = make_response(devis_rh(demande))
    response.headers["content-type"] = "application/pdf"
    content_disposition = 'attachment;filename="devis-rh.pdf"'
    response.headers["content-disposition"] = content_disposition.encode()
    return response


@route("/calculettes/devis_rh", methods=["POST"])
def calculette_rh(db: SQLAlchemy, structure_repo: StructureRepository):
    demande = DemandeRH()

    json = request.json

    model = json["model"]
    form = json["form"]

    user = get_current_profile()

    model = cleanup_model(model, form)

    demande = DemandeRH(data=model)
    demande.form_state = form

    porteur_dto = model.get("porteur")
    if porteur_dto:
        porteur_id = porteur_dto["value"]
        demande.porteur = db.session.query(Profile).get(porteur_id)

    if user != demande.porteur:
        demande.gestionnaire = user

    structure_dto = model.get("laboratoire")
    if structure_dto:
        structure_id = structure_dto["value"]
        demande.structure = structure_repo.get_by_id(structure_id)

    blob = devis_rh(demande)
    response = make_response(blob)
    response.headers["content-type"] = "application/pdf"
    return response


@route("/demandes/<int:id>/lettre_commande_rh")
def lettre_commande_rh_rest(id, db: SQLAlchemy):
    demande = db.session.query(Demande).get(id)
    if acces_restreint(demande):
        raise Forbidden()

    assert isinstance(demande, DemandeRH)

    response = make_response(lettre_commande_rh(demande))
    response.headers["content-type"] = "application/pdf"
    content_disposition = 'attachment;filename="lettre-commande-rh.pdf"'
    response.headers["content-disposition"] = content_disposition.encode()
    return response


@route("/demandes/<int:id>/fdc")
def fdc(id: int, db: SQLAlchemy):
    demande = db.session.query(Demande).get(id)

    # model = demande.feuille_cout
    html = demande.data["fdc_html"].encode("utf8")

    return html, 200, {"content-type": "text/html"}

    # css = CSS(string="")
    #
    # s = HTML(string=html).write_pdf(stylesheets=[css])
    # assert s
    # return html, 200, {"content-type": "application/pdf"}

    # response = make_response(s)
    # response.headers["content-type"] = "application/pdf"
    # return response


@route("/demandes_a_valider/<type>")
@route("/demandes_a_valider/")
def demandes_a_valider(type=None):
    if type:
        return redirect(f"/#/demandes_a_valider/{type}")
    else:
        return redirect("/#/tasks")



from __future__ import annotations

from functools import wraps
from typing import Union

import structlog
from flask import g, redirect, session, url_for
from flask_login import AnonymousUserMixin, UserMixin

from labster.auth import AuthContext
from labster.di import injector
from labster.domain2.model.profile import Profile, ProfileRepository


def login_required(func):
    @wraps(func)
    def decorated_view(*args, **kwargs):
        current_user = get_current_user()
        if not current_user.is_authenticated:
            return redirect(url_for("auth.login", _external=True))

        return func(*args, **kwargs)

    return decorated_view


class AnonymousUser(AnonymousUserMixin):
    name = "Anonymous"

    def has_role(self, *args, **kw):
        return False


class AuthenticatedUser(UserMixin):
    profile: Profile

    def __init__(self, profile):
        self.profile = profile

    @property
    def name(self):
        return self.profile.name

    @property
    def uid(self):
        return self.profile.uid


User = Union[AuthenticatedUser, AnonymousUser]


def login_user():
    current_user_id = session.get("current_user_id", "ANONYMOUS")
    logger = structlog.get_logger()

    if current_user_id == "ANONYMOUS":
        g.current_user = AnonymousUser()
        g.current_profile = None
        logger.new(user=None)

    else:
        try:
            profile_repos = injector.get(ProfileRepository)
            profile = profile_repos.get_by_id(current_user_id)

            g.current_profile = profile
            g.current_user = AuthenticatedUser(profile)

        except:
            g.current_user = AnonymousUser()
            g.current_profile = None

        logger.new(user=g.current_user.name)


def get_current_user() -> User:
    auth_context = injector.get(AuthContext)
    user = auth_context.current_user
    assert user
    return user


def get_current_profile() -> Profile:
    auth_context = injector.get(AuthContext)
    profile = auth_context.current_profile
    assert profile
    return profile



from __future__ import annotations

import traceback
from datetime import date, timedelta

import ramda as r
from flask_sqlalchemy import SQLAlchemy
from jsonrpcserver import method
from sqlalchemy import or_
from sqlalchemy.orm import joinedload

from labster.di import injector
from labster.domain2.model.demande import Demande, DemandeAutre, \
    DemandeAvenantConvention, DemandeConvention, DemandePiMixin, DemandeRH
from labster.domain2.model.profile import Profile
from labster.domain2.services.contacts import ContactService
from labster.domain2.services.roles import Role, RoleService
from labster.domain2.services.workflow.states import EN_EDITION, EN_VALIDATION
from labster.rbac import get_drv_membership, is_membre_dri, is_membre_drv
from labster.rpc.queries.cache import memoize
from labster.rpc.util import owner_sorter
from labster.security import get_current_profile
from labster.types import JSON, JSONDict, JSONList
from labster.util import url_for


def base_query():
    db = injector.get(SQLAlchemy)

    return db.session.query(Demande).options(
        joinedload(Demande.structure),
        joinedload(Demande.contact_labco),
        joinedload(Demande.gestionnaire),
        joinedload(Demande.porteur),
    )


@method
def get_demandes(scope="all", archives=False, tag="") -> JSONList:
    archives = bool(archives)
    profile = get_current_profile()

    view = get_table_view(scope, profile, archives)
    if not view:
        return []

    demandes: list[Demande] = view.get_demandes_for(profile)

    def make_pred(tag):
        def pred(demande):
            if tag == "rh":
                return isinstance(demande, DemandeRH)
            elif tag == "conventions":
                return isinstance(
                    demande, (DemandeConvention, DemandeAvenantConvention)
                )
            elif tag == "pi":
                return isinstance(demande, DemandePiMixin)
            elif tag == "autres":
                return isinstance(demande, DemandeAutre)
            else:
                raise RuntimeError()

        return pred

    if tag:
        demandes = r.filter(make_pred(tag), demandes)

    demandes.sort(key=lambda d: d.created_at, reverse=True)
    return demandes_to_json(demandes, profile)


def get_table_view(
    scope: str, user: Profile, archives: bool = False
) -> TableView | None:
    all_table_views = [
        cls()
        for cls in globals().values()
        if isinstance(cls, type) and issubclass(cls, TableView)
    ]

    view = None
    for x in all_table_views:
        if x.scope != scope or x.archives != archives:
            continue

        view = x
        break

    if not view:
        return None

    if not view.is_visible_for(user):
        return None

    return view


#
# Demandes actives
#
class TableView:
    scope = ""
    title = ""
    archives = False

    def is_visible_for(self, user: Profile):
        return False

    def get_demandes_for(self, user: Profile):
        return []


#
# Scopes génériques
#
class MesDemandesTableView(TableView):
    scope = "mes demandes"

    def is_visible_for(self, user: Profile):
        return True

    def get_demandes_for(self, user: Profile):
        return mes_demandes(user)


class MesTachesTableView(TableView):
    scope = "mes tâches"

    def is_visible_for(self, user: Profile):
        return True

    def get_demandes_for(self, user: Profile):
        return mes_taches(user)


class MesDemandesEnRetardTableView(TableView):
    scope = "mes demandes en retard"

    def is_visible_for(self, user: Profile):
        return True

    def get_demandes_for(self, user: Profile):
        return mes_taches_en_retard(user)


class DemandesAValiderTableView(TableView):
    scope = "demandes à valider"

    def is_visible_for(self, user: Profile):
        return True

    def get_demandes_for(self, user: Profile):
        demandes = mes_taches(user)
        demandes = [d for d in demandes if d.wf_state == EN_VALIDATION.id]
        return demandes


def mes_demandes(
    user: Profile, archived: bool = False, all: bool = False
) -> list[Demande]:
    query = (
        base_query()
        .filter(
            or_(
                (Demande.porteur == user),
                (Demande.contact_labco == user),
                (Demande.gestionnaire == user),
            )
        )
        .filter_by(active=(not archived))
        .order_by(Demande.created_at.desc())
    )
    return query.all()


@memoize()
def mes_taches(user: Profile) -> list[Demande]:
    """Retourne la liste des demandes pour lesquels l'utilisateur a une action
    à réaliser."""
    query = (
        base_query().filter(Demande.active == True).order_by(Demande.created_at.desc())
    )

    demandes = query.all()

    def is_task(demande):
        workflow = demande.get_workflow(user)
        state = workflow.current_state()
        owners = state.task_owners(workflow)
        return user.id in (o.id for o in owners)

    demandes = [d for d in demandes if is_task(d)]
    return demandes


def mes_taches_en_retard(user: Profile) -> list[Demande]:
    demandes = mes_taches(user)

    def en_retard(d) -> bool:
        return (d.wf_retard or 0) > 0

    demandes = list(filter(en_retard, demandes))
    demandes = sorted(demandes, key=lambda d: d.wf_retard, reverse=True)
    return demandes


#
# Scopes "recherche"
#
class PorteurTableView(TableView):
    scope = "porteur"
    title = "Demandes dont je suis porteur"

    def is_visible_for(self, user: Profile):
        return user.has_role(Role.PORTEUR, "*")

    def get_demandes_for(self, user: Profile):
        demandes_comme_porteur = (
            base_query()
            .filter(Demande.porteur == user)
            .filter_by(active=(not self.archives))
            .order_by(Demande.created_at.desc())
            .all()
        )
        toutes_les_demandes = (
            base_query()
            .filter_by(active=(not self.archives))
            .order_by(Demande.created_at.desc())
            .all()
        )

        demandes_comme_contributeur = []
        for demande in toutes_les_demandes:
            contributeurs_raw = demande.data.get("contributeurs", [])
            for dd in contributeurs_raw:
                contributeur_id = dd["value"]
                if contributeur_id == user.id:
                    demandes_comme_contributeur += [demande]

        def sorter(d):
            return d.created_at

        return sorted(
            set(demandes_comme_porteur + demandes_comme_contributeur), key=sorter
        )


class GestionnaireTableView(TableView):
    scope = "gestionnaire"
    title = "Demandes dont je suis gestionnaire"

    def is_visible_for(self, user: Profile):
        return user.has_role(Role.GESTIONNAIRE, "*")

    def get_demandes_for(self, user: Profile):
        return (
            base_query()
            .filter(Demande.gestionnaire == user)
            .filter_by(active=(not self.archives))
            .order_by(Demande.created_at.desc())
            .all()
        )


class MesStructuresTableView(TableView):
    scope = "mes structures"
    title = "Demandes des structures dont je suis un responsable/gestionnaire"

    def is_visible_for(self, user: Profile):
        return user.has_role(Role.GESTIONNAIRE, "*") or user.has_role(
            Role.RESPONSABLE, "*"
        )

    def get_demandes_for(self, user: Profile):
        role_service = injector.get(RoleService)
        roles = role_service.get_roles_for_user(user)
        structures = set(roles[Role.GESTIONNAIRE]) | set(roles[Role.RESPONSABLE])
        for s in set(structures):
            structures |= s.descendants

        structure_ids = {s.id for s in structures}

        toutes_les_demandes = (
            base_query()
            .filter_by(active=(not self.archives))
            .order_by(Demande.created_at.desc())
            .all()
        )

        result = []
        for demande in toutes_les_demandes:
            if demande.structure_id in structure_ids:
                result.append(demande)
                continue

            for d in demande.data.get("structures_concernees", []):
                structure_id = d["value"]
                if structure_id in structure_ids:
                    result.append(demande)

        return result


#
# Archives recherche
#
class ArchivesPorteurTableView(PorteurTableView):
    title = "Demandes archivées dont j'ai été porteur"
    archives = True


class ArchivesGestionnaireTableView(GestionnaireTableView):
    title = "Demandes archivées dont j'ai été gestionnaire"
    archives = True


class ArchivesMesStructuresTableView(MesStructuresTableView):
    title = "Demandes archivées des structures dont je suis un responsable/gestionnaire"
    archives = True


#
# DRI et DRV
#
class DriTableView(TableView):
    scope = "dri"
    title = "Demandes en cours à la DR&I et dans les DRV"

    def is_visible_for(self, user: Profile):
        return is_membre_dri(user)

    def get_demandes_for(self, user: Profile):
        query = (
            base_query()
            .filter_by(active=(not self.archives))
            .filter(Demande.contact_labco != None)
            .order_by(Demande.created_at.desc())
        )
        return filter_on_state_for_dri(query.all())


class DrvTableView(TableView):
    scope = "drv"
    title = "Demandes en cours dans ma DRV"

    def is_visible_for(self, user: Profile):
        return is_membre_drv(user)

    def get_demandes_for(self, user: Profile):
        query = (
            base_query()
            .filter_by(active=(not self.archives))
            .filter(Demande.contact_labco != None)
            .order_by(Demande.created_at.desc())
        )
        demandes = query.all()

        drv = get_drv_membership(user)
        if not drv:
            return []

        fac = drv.parent
        assert fac
        structures_possibles = {fac} | fac.descendants
        ids_structures_possibles = {s.id for s in structures_possibles}

        demandes = [
            demande
            for demande in demandes
            if demande.structure and demande.structure.id in ids_structures_possibles
        ]
        return filter_on_state_for_dri(demandes)


class ContactTableView(TableView):
    scope = "contact"
    title = "Demandes dont je suis contact"

    def is_visible_for(self, user: Profile):
        return is_membre_dri(user) or is_membre_drv(user)

    def get_demandes_for(self, user: Profile):
        return filter_on_state_for_dri(
            base_query()
            .filter(Demande.contact_labco == user)
            .filter_by(active=(not self.archives))
            .order_by(Demande.created_at.desc())
            .all()
        )


@memoize()
def get_mapping():
    contact_service = injector.get(ContactService)
    return contact_service.get_mapping()


class MesStructuresDriOuDrvTableView(TableView):
    scope = "mes structures dri"
    title = "Demandes des structures dont je suis un contact Lab&Co"

    def is_visible_for(self, user: Profile):
        mapping = get_mapping()
        return (is_membre_dri(user) or is_membre_drv(user)) and mapping

    def get_demandes_for(self, user: Profile):
        mapping = get_mapping()
        structures = {s for s, d in mapping.items() if user in d.values()}
        structure_ids = {s.id for s in structures}
        return filter_on_state_for_dri(
            base_query()
            .filter(Demande.structure_id.in_(structure_ids))
            .filter(Demande.contact_labco != None)
            .filter_by(active=(not self.archives))
            .order_by(Demande.created_at.desc())
            .all()
        )


#
# Archives DR&I
#
class ArchivesDriTableView(DriTableView):
    title = "Demandes archivées à la DR&I"
    archives = True


class ArchivesDrvTableView(DrvTableView):
    title = "Demandes archivées dans ma DRV"
    archives = True


class ArchivesContactTableView(ContactTableView):
    title = "Demandes archivées dont j'ai été le contact"
    archives = True


class ArchivesMesStructuresDriOuDrvTableView(MesStructuresDriOuDrvTableView):
    title = "Demandes archivées des structures dont je suis un contact"
    archives = True


#
# Util
#
def filter_on_state_for_dri(demandes: list[Demande]) -> list[Demande]:
    result = []
    for demande in demandes:
        wf_states = {
            state["new_state"] for state in demande.wf_history if "new_state" in state
        }

        if wf_states != {EN_EDITION.id}:
            result.append(demande)

    return result


#
# Serialization
#
def demandes_to_json(demandes: list[Demande], user: Profile) -> JSONList:
    result: list[JSON] = []

    for demande in demandes:
        try:
            row = demande_to_json(demande, user)
            result.append(row)
        except Exception:
            traceback.print_exc()

    return result


def demande_to_json(demande: Demande, user: Profile) -> JSONDict:
    row: JSONDict = {
        "created_at": format_date(demande.created_at),
        "__created_at__": isoformat_date(demande.created_at),
        "icon_class": demande.icon_class,
        "type": demande.type,
        "url": url_for(demande),
        "nom": demande.nom or demande.titre,
        "age": demande.age,
        "date_debut": format_date(demande.date_debut),
        "__date_debut__": isoformat_date(demande.date_debut),
    }

    date_soumission = demande.date_soumission
    if date_soumission:
        duree_traitement: timedelta = date.today() - date_soumission
        row["date_soumission"] = (
            format_date(date_soumission) + f" ({duree_traitement.days}j)"
        )
        row["__date_soumission__"] = isoformat_date(date_soumission)
    else:
        row["date_soumission"] = ""
        row["__date_soumission__"] = ""

    row["retard"] = demande.retard
    row["no_infolab"] = demande.no_infolab

    for k in ("porteur", "gestionnaire", "contact_labco"):
        profile = getattr(demande, k)
        if profile:
            row[k] = profile.full_name
            row[k + "_url"] = url_for(profile)
            nom = getattr(profile, "nom", "")
            prenom = getattr(profile, "prenom", "")
            # helps sorting in JS
            row[k + "_nom"] = f"{nom}, {prenom}"
        else:
            row[k] = ""
            row[k + "_url"] = ""
            row[k + "_nom"] = ""

    structure = demande.structure
    if structure:
        row["structure"] = structure.sigle_ou_nom
        row["structure_url"] = url_for(structure)
        # laboratoire = structure.laboratoire
        # row["laboratoire"] = laboratoire.sigle_ou_nom
        # row["laboratoire_url"] = url_for(laboratoire)
    else:
        row["structure"] = ""
        row["structure_url"] = ""
    # TODO: remove
    row["laboratoire"] = ""
    row["laboratoire_url"] = ""

    # Workflow
    workflow = demande.get_workflow(user)
    state = demande.get_state()
    row["etat"] = state.label_short
    row["prochaine_action"] = state.next_action
    owners = list(workflow.current_owners())
    owners.sort(key=owner_sorter)
    if owners:
        row["owner"] = f"{owners[0].nom}, {owners[0].prenom}"
    else:
        row["owner"] = ""

    row["owners"] = [
        {
            "name": f"{owner.nom}, {owner.prenom}",
            "id": owner.id,
        }
        for owner in owners
    ]

    return row


def format_date(d: date | None) -> str:
    if d:
        return d.strftime("%d/%m/%Y")
    else:
        return ""


def isoformat_date(d: date | None) -> str:
    if d:
        return d.isoformat()
    else:
        return ""



# from __future__ import annotations
#
# from sqlalchemy import Column, Date, Integer, String, Unicode
#
# from labster.extensions import db
#
#
# # class StatsLine(db.Model):
# #     __tablename__ = "stats_line"
# #
# #     demande_id = Column(Integer, primary_key=True)
# #
# #     # Filtrage
# #     type_demande = Column(String, index=True, nullable=False)
# #
# #     date_soumission = Column(Date, index=True)
# #     date_finalisation = Column(Date, index=True)
# #
# #     structure_id = Column(Integer, index=True)
# #     porteur_id = Column(Integer, index=True)
# #
# #     # Champs spécifiques
# #     financeur = Column(Unicode, index=True)
# #     type_recrutement = Column(Unicode, index=True)
# #
# #     wf_state = Column(Unicode, index=True)
# #
# #     # Drilldown
# #     pole_id = Column(Integer, index=True)
# #     ufr_id = Column(Integer, index=True)
# #     labo_id = Column(Integer, index=True)
# #     departement_id = Column(Integer, index=True)
# #     equipe_id = Column(Integer, index=True)
# #
# #     # Agregations / Demande Conventions
# #     montant = Column(Integer)
# #     recrutements_prev = Column(Integer)
# #
# #     # Agregations / Demandes RH
# #     salaire_brut_mensuel = Column(Integer)
# #     cout_total_mensuel = Column(Integer)
# #
# #     # Commun Conventions / RH
# #     duree = Column(Integer)
# #
# #     # Autres
# #     duree_traitement = Column(Integer)
from __future__ import annotations



from __future__ import annotations

from functools import singledispatch

from tqdm import tqdm

from labster.domain2.model.demande import Demande, DemandeAvenantConvention, \
    DemandeConvention, DemandeRH
from labster.domain2.services.calculs_couts import get_ctx_for_demande
from labster.extensions import db

from .model import StatsLine


def sync_all():
    StatsLine.query.delete()

    demandes = Demande.query.all()

    for demande in tqdm(demandes, disable=None):
        try:
            line = sync(demande)
        except ValueError:
            print(f"Error on demande: {demande.id}")
            line = None

        if line:
            db.session.add(line)

    db.session.commit()


def sync(demande):
    if demande.wf_state == "EN_EDITION":
        keep = False
        for state in demande.wf_history:
            transition = state.get("transition")
            if transition in (
                "REQUERIR_MODIFICATION_DIR",
                "REQUERIR_MODIFICATION_DGRTT",
            ):
                keep = True
        if not keep:
            return

    data = {
        "demande_id": demande.id,
        "type_demande": demande.type,
        "wf_state": demande.wf_state,
        "date_soumission": demande.date_soumission,
        "date_finalisation": demande.date_finalisation,
    }

    if demande.porteur:
        data["porteur_id"] = demande.porteur_id
    else:
        data["porteur_id"] = None

    structure = demande.structure
    if structure:
        data["structure_id"] = structure.id
        path = list(reversed(structure.ancestors)) + [structure]
        for i, node in enumerate(path):
            data[f"l{i+1}"] = node.id

    update_data = sync_specific(demande)

    data.update(update_data)
    line = StatsLine(**data)
    return line


@singledispatch
def sync_specific(demande):
    data = {}
    return data


@sync_specific.register(DemandeConvention)
def _(demande):
    data = {
        "montant": pick_integer(demande, "montant_financement"),
        "recrutements_prev": pick_integer(demande, "nombre_recrutements"),
        "duree": pick_integer(demande, "duree_previsionnelle"),
        "financeur": getattr(demande, "type_financeur", ""),
    }
    return data


@sync_specific.register(DemandeAvenantConvention)
def _1(demande):
    data = {"financeur": getattr(demande, "type_financeur", "")}
    return data


@sync_specific.register(DemandeRH)
def _2(demande):
    # print(demande.id)
    data = {
        "duree": demande.duree_mois or 0,
        "type_recrutement": (
            getattr(demande, "nature_du_recrutement", "")
            + " | "
            + getattr(demande, "type_de_demande", "")
        ),
    }

    ctx = get_ctx_for_demande(demande)
    data["salaire_brut_mensuel"] = ctx["salaire_brut"]
    data["cout_total_mensuel"] = ctx["cout_total"]

    return data


def pick_integer(demande, field_name):
    try:
        value = int(getattr(demande, field_name).replace(" ", ""))
        if value > 2147483647:
            value = 2147483647
    except (ValueError, AttributeError):
        # print(getattr(demande, field_name, None))
        # traceback.print_exc()
        value = 0
    return value


def get_id(obj):
    if obj:
        return obj.id
    else:
        return None



from __future__ import annotations


def get_form(
    require_note=False,
    ask_for_revalidation=False,
    with_no_infolab=False,
    with_no_eotp=False,
):
    form = [
        {"name": "note", "label": "Note", "type": "textarea", "required": require_note},
    ]

    if ask_for_revalidation:
        form.append(
            {
                "name": "resoumission",
                "label": "Redemander validation par la hiérarchie",
                "type": "bool",
                "required": False,
            },
        )

    if with_no_infolab:
        form.append(
            {
                "name": "no_infolab",
                "label": "N° Infolab / Référence",
                "type": "text",
                "required": True,
            },
        )
    if with_no_eotp:
        form.append(
            {"name": "no_eotp", "label": "N° EOTP", "type": "text", "required": True},
        )

    return form



from __future__ import annotations

from typing import Any

from jsonrpcserver import method

from labster.di import injector
from labster.domain2.model.profile import Profile, ProfileRepository
from labster.domain2.model.structure import Structure, StructureRepository
from labster.domain2.services.contacts import ContactService, ContactType
from labster.domain2.services.roles import Role, RoleService
from labster.ldap.constants import DRI_ET_DRV_DNS
from labster.security import get_current_profile
from labster.types import JSONDict, JSONList

structure_repo = injector.get(StructureRepository)
profile_repo = injector.get(ProfileRepository)
contact_service = injector.get(ContactService)
role_service = injector.get(RoleService)


@method
def get_contacts(structure_id) -> JSONList:
    structure = structure_repo.get_by_id(structure_id)
    assert structure
    return make_contacts_dto(structure)


@method
def get_all_contacts() -> JSONList:
    structures: list[Structure] = list(structure_repo.get_all())
    structures.sort(key=lambda x: x.nom)

    result = []
    for structure in structures:
        for contact_type in ContactType:
            contact = contact_service.get_contact(structure, contact_type)
            if contact:
                structure_dto = {
                    "name": structure.nom,
                    "id": structure.id,
                    "type": structure.type_name,
                    "sigle": structure.sigle,
                }
                result.append(
                    {
                        "structure": structure_dto,
                        "contacts": make_contacts_dto(structure),
                    }
                )
                break

    return result


@method
def get_membres_dri() -> JSONList:
    membres_dri_et_drv_set = set()
    for dn in DRI_ET_DRV_DNS:
        structure = structure_repo.get_by_dn(dn)
        membres_structure = role_service.get_users_with_given_role(
            Role.MEMBRE, structure
        )
        membres_dri_et_drv_set.update(membres_structure)

    membres_dri_et_drv = list(membres_dri_et_drv_set)
    membres_dri_et_drv.sort(key=lambda x: (x.nom, x.prenom))
    return [
        {"uid": m.uid, "nom": m.nom, "prenom": m.prenom} for m in membres_dri_et_drv
    ]


@method
def get_contacts_for_user() -> JSONDict:
    user = get_current_profile()

    roles: dict[Role, set[Structure]] = role_service.get_roles_for_user(user)
    structures = list(roles[Role.MEMBRE])
    structures.sort(key=lambda x: x.nom)

    result1 = []
    for structure in structures:
        for contact_type in ContactType:
            contact = contact_service.get_contact(structure, contact_type)
            if contact:
                structure_dto = {
                    "name": structure.nom,
                    "id": structure.id,
                    "type": structure.type_name,
                    "sigle": structure.sigle,
                }
                result1.append(
                    {
                        "structure": structure_dto,
                        "contacts": make_contacts_dto(structure),
                    }
                )
                break

    result2 = []
    mapping: dict[Structure, dict[ContactType, Profile]] = contact_service.get_mapping()
    for structure, d in mapping.items():
        if user in d.values():
            for contact_type, user1 in d.items():
                if user1 == user:
                    structure_dto = {
                        "name": structure.nom,
                        "id": structure.id,
                        "type": structure.type_name,
                        "sigle": structure.sigle,
                    }
                    result2.append(
                        {
                            "structure": structure_dto,
                            "bureau": contact_type.value,
                        }
                    )

    return {
        "structures": result1,
        "mes_contacts": result2,
    }


#
# Serialization
#
def make_contacts_dto(structure: Structure) -> list[dict[str, Any]]:
    result = []
    for contact_type in ContactType:
        contact = contact_service.get_contact(structure, contact_type)
        if contact:
            dto = {
                "id": contact.id,
                "uid": contact.uid,
                "name": contact.full_name,
                "tel": contact.telephone,
                "email": contact.email,
                "type_name": contact_type.name,
                "type_value": contact_type.value,
            }
        else:
            dto = {
                "id": None,
                "uid": None,
                "name": "",
                "type_name": contact_type.name,
                "type_value": contact_type.value,
            }
        result.append(dto)
    return result



from __future__ import annotations

import re
from datetime import datetime
from pathlib import Path

from abilian.core.models.blob import Blob
from flask import current_app
from flask_sqlalchemy import SQLAlchemy
from jsonrpcserver import method
from sqlalchemy import func
from toolz import first
from weasyprint import CSS, HTML
from werkzeug.exceptions import BadRequest, Forbidden, NotFound

from labster.di import injector
from labster.domain2.model.demande import Demande, demande_factory
from labster.domain2.model.profile import Profile
from labster.domain2.model.structure import Structure, StructureRepository
from labster.extensions import whoosh
from labster.rbac import check_can_add_pj, check_read_access, \
    check_write_access, feuille_cout_editable
from labster.security import get_current_profile

db = injector.get(SQLAlchemy)
structure_repo = injector.get(StructureRepository)


PRINT_CSS = """
.print .feuille-cout {
    input {
        display: none;
    }

    select {
        display: none;
    }

    .no-print {
        display: none;
    }

    .only-print {
        display: inline;
    }
}

.only-print {
    display: none;
}
"""


@method
def create_demande(model, form):
    user = get_current_profile()
    # TODO: check permissions

    model = cleanup_model(model, form)

    form_type = form["name"]

    demande = demande_factory(type=form_type, demandeur=user, data=model)

    demande.porteur = get_porteur(model)
    demande.structure = get_structure(model)

    if user != demande.porteur:
        demande.gestionnaire = user
    else:
        demande.gestionnaire = None

    demande.form_state = form

    new_id = db.session.query(func.max(Demande.id)).one()[0] + 1
    demande.id = new_id

    db.session.add(demande)
    db.session.commit()

    messages = [
        ["Votre demande a été créée.", "success"],
    ]
    if not demande.is_valid():
        messages.append(["Attention, votre demande est encore incomplète.", "warning"])

    whoosh.index_object(demande)

    return {
        "id": demande.id,
        "messages": messages,
    }


@method
def update_demande(id, model, form):
    user = get_current_profile()
    model = cleanup_model(model, form)

    demande = db.session.query(Demande).get(id)
    check_read_access(demande)
    check_write_access(demande)

    demande.form_state = form

    messages = []
    if not demande.has_same_data(model):
        demande.update_data(model)
        porteur = get_porteur(model)
        if porteur != demande.porteur:
            demande.porteur = porteur
            demande.gestionnaire = user
        demande.structure = get_structure(model)
        db.session.commit()

        messages.append(
            [
                "Votre demande a été modifiée. La version précédente de la demande a été archivée.",
                "success",
            ]
        )
    else:
        messages.append(["Vous n'avez pas effectué de modification.", "success"])

    if not demande.is_valid():
        messages.append(["Attention, votre demande est encore incomplète.", "warning"])

    whoosh.index_object(demande)

    return messages


def get_structure(model: dict) -> Structure | None:
    structure_dto = model.get("laboratoire")
    if not structure_dto:
        return None

    if isinstance(structure_dto, str):
        structure_id = structure_dto
    elif isinstance(structure_dto, dict):
        structure_id = structure_dto["value"]
    else:
        raise BadRequest

    structure = structure_repo.get_by_id(structure_id)
    return structure


def get_porteur(model: dict) -> Profile | None:
    porteur_dto = model.get("porteur")
    if not porteur_dto:
        return None

    if isinstance(porteur_dto, str):
        porteur_id = porteur_dto
    elif isinstance(porteur_dto, dict):
        porteur_id = porteur_dto["value"]
    else:
        raise BadRequest
    porteur = db.session.query(Profile).get(porteur_id)
    return porteur


@method
def dupliquer_demande(id):
    demande = db.session.query(Demande).get(id)
    check_read_access(demande)

    new_id = db.session.query(func.max(Demande.id)).one()[0] + 1
    nouvelle_demande = demande.clone()
    nouvelle_demande.id = new_id

    db.session.add(nouvelle_demande)
    db.session.commit()

    return nouvelle_demande.id


@method
def delete_pj(demande_id, blob_id):
    demande: Demande = db.session.query(Demande).get(demande_id)
    check_can_add_pj(demande)

    attachments = demande.attachments.copy()
    filename = None
    for filename, v in demande.attachments.items():
        if v["id"] == blob_id:
            filename = filename
            break
    if filename:
        del attachments[filename]
    demande.attachments = attachments
    db.session.commit()


@method
def update_feuille_de_cout(model, html):
    id = model["id"]
    demande = db.session.query(Demande).get(id)

    if not demande:
        raise NotFound()

    if not feuille_cout_editable(demande):
        raise Forbidden()

    demande.feuille_cout = model
    html = re.sub('id="feuille-cout"', 'id="feuille-cout" class="print"', html)
    demande.data["fdc_html"] = f'<!DOCTYPE html><html lang="fr-FR">{html}</html>'
    db.session.commit()

    root = Path(current_app.root_path)
    static = root / "static"
    css = static / "css"

    css1 = CSS(filename=str(first(css.glob("chunk*"))))
    css2 = CSS(filename=str(first(css.glob("app*"))))
    css3 = CSS(string=PRINT_CSS)

    s = HTML(string=html).write_pdf(stylesheets=[css3, css1, css2, css3])
    assert s

    blob = Blob(s)
    db.session.add(blob)
    db.session.flush()

    d = {
        "name": "Feuille de coût.pdf",
        "date": datetime.utcnow().isoformat(),
        "blob_id": blob.id,
    }
    if demande.documents_generes:
        docs = list(demande.documents_generes)
        docs.append(d)
        demande.documents_generes = docs
    else:
        demande.documents_generes = [d]
    db.session.commit()


def make_pdf(model):
    pass


#
# Utility functions
#
def cleanup_model(model, form):
    """Remove values for fields that are not visible."""
    new_model = {}
    fields = form["fields"]
    for key, value in model.items():
        if key not in fields:
            continue
        if key.startswith("html-"):
            continue
        field = fields[key]
        if not field.get("visible") and value:
            value = None
        new_model[key] = value

    return new_model



from __future__ import annotations

from flask import Flask
from jsonrpcserver import method
from werkzeug.exceptions import Unauthorized

from labster.blueprints.auth.cas import get_user_by_login, login_user
from labster.di import injector


@method
def backdoor(login="poulainm"):
    app = injector.get(Flask)

    if not (app.config.get("ALLOW_BACKDOOR") or app.testing):
        raise Unauthorized()

    user = get_user_by_login(login)
    login_user(user)

    return None



""""""
from __future__ import annotations

import os
import shutil
import sys
from pathlib import Path
from pprint import pprint

import click
import structlog
from abilian.services import audit_service, index_service
from flask import current_app
from flask.cli import with_appcontext
from flask_sqlalchemy import SQLAlchemy
from pampy import match

from labster.bi.synchro import sync_all
from labster.blueprints.notifications.mails import send_notification_to, \
    send_recap_to
from labster.di import injector
from labster.domain2.model.demande import Demande
from labster.domain2.model.profile import DAILY, WEEKLY, Profile, \
    ProfileRepository
from labster.domain2.model.structure import StructureRepository
from labster.domain2.services.roles import Role, RoleService
from labster.ldap import sync
from labster.rpc.cache import cache

logger = structlog.get_logger()

profile_repository = injector.get(ProfileRepository)
db = injector.get(SQLAlchemy)
role_service = injector.get(RoleService)


#
# Development commands
#
@click.command()
@click.option("--server", default="flask")
@click.option("--coverage")
@with_appcontext
def devserver(server, coverage=False):
    if server == "flask":
        if coverage:
            flask = shutil.which("flask")
            web_cmd = f"coverage run {flask} run --no-reload"
        else:
            web_cmd = "flask run"
    elif server == "gunicorn":
        web_cmd = "gunicorn -w1 --timeout 300 --bind 0.0.0.0:5000 'wsgi:app'"
    elif server == "uvicorn":
        web_cmd = "uvicorn --host 0.0.0.0 --port 5000 --wsgi 'wsgi:app'"
    else:
        raise click.BadParameter("No such server")

    daemons = [("web", web_cmd), ("webpack", "yarn --cwd front run serve")]
    run_daemons(daemons)


def run_daemons(daemons):
    import honcho.manager

    root = (Path(current_app.root_path) / "..").resolve()
    if not (root / ".git").exists():
        print(f"root = {root}")
        print("project must be installed in development mode")
        sys.exit(1)

    manager = honcho.manager.Manager()
    for name, cmd in daemons:
        manager.add_process(name, cmd, cwd=str(root))

    manager.loop()
    sys.exit(manager.returncode)


@click.command()
@with_appcontext
def config():
    config = current_app.config
    print("Current config:")
    print()
    pprint(dict(config))
    print("Current env:")
    pprint(dict(os.environ))


@click.command()
@with_appcontext
def dump_structures():
    structure_repo = injector.get(StructureRepository)
    structures = structure_repo.get_all()
    for s in structures:
        print(s.nom)
        print("Children:", s.children)
        print("Parents:", s.parents)
        print()


#
# Production commands
#
@click.command()
@with_appcontext
def create_tables():
    print("# Creating tables")
    db.create_all()


@click.command()
@with_appcontext
def create_new_tables():
    print("# Creating new tables")

    table_names = db.metadata.tables.keys()
    for name in table_names:
        if name.startswith("v3_"):
            print(f"Dropping {name}")
            db.engine.execute(f'drop table if exists "{name}" cascade;')

    db.create_all()


@click.command()
@with_appcontext
def drop_all_tables():
    print("# Dropping all tables")

    table_names = db.metadata.tables.keys()
    for name in table_names:
        print(f"Dropping {name}")
        db.engine.execute(f'drop table if exists "{name}" cascade;')

    db.create_all()


@click.command()
@with_appcontext
def ldap_sync():
    print("# Synchro LDAP")
    audit_service.stop(ignore_state=True)
    index_service.stop(ignore_state=True)

    sync.sync_users()

    db.session.commit()
    cache.clear()


@click.command()
@click.option("--max", default=0)
@with_appcontext
def update_roles(max):
    print("# Updating roles")
    audit_service.stop(ignore_state=True)
    index_service.stop(ignore_state=True)

    sync.update_roles(max)

    db.session.commit()


@click.command()
@with_appcontext
def show_roles():
    users = list(profile_repository.get_all())
    users.sort(key=lambda x: (x.nom, x.prenom))
    for user in users:
        print(user)
        roles = role_service.get_roles_for_user(user)
        pprint(dict(roles))
        print()


@click.command()
@with_appcontext
def syncbi():
    print("# updating syncing BI data")
    sync_all()


@click.command()
@with_appcontext
def update_retard():
    print("# updating 'retards'")
    demandes = Demande.query.all()
    for demande in demandes:
        demande.update_retard()

    db.session.commit()


@click.command()
@click.option("-v", "--verbose", is_flag=True)
@with_appcontext
def send_recap(verbose):
    users = Profile.query.filter_by(active=True).order_by(Profile.nom).all()
    for user in users:
        status = send_recap_to(user)
        if status and verbose:
            print(f"recap sent to {user.login}")

    db.session.commit()


@click.command()
@click.argument("frequency", required=True, type=str)
@click.option("-v", "--verbose", is_flag=True)
@with_appcontext
def send_notifications(frequency, verbose=False):
    # fmt: off
    preference = match(
        frequency,
        "daily", DAILY,
        "weekly", WEEKLY,
    )
    # fmt: on
    users = (
        Profile.query.filter(Profile.active == True)
        .filter(Profile.preferences_notifications == preference)
        .order_by(Profile.nom)
        .all()
    )

    for user in users:
        if verbose:
            print(f"Sending notification to {user.login}")
        send_notification_to(user)

    db.session.commit()


@click.command()
@with_appcontext
def grant_mnp():
    user = db.session.query(Profile).filter(Profile.login == "poulainm").one()
    role_service.grant_role(user, Role.ADMIN_CENTRAL)

    db.session.commit()


@click.command()
@with_appcontext
def fix():
    # from .fixes import fix_ajout_supann_code_entite
    # fix_ajout_supann_code_entite()

    print("No fix needed")



from __future__ import annotations

from labster.domain2.model.profile import Profile


def test_1():
    profile = Profile(nom="NOM", prenom="Prenom")
    assert profile.full_name == "Prenom NOM"


def test_2():
    profile = Profile()
    assert profile.prenom == ""
    assert profile.nom == ""



from __future__ import annotations

import toolz
from abilian.web.search.views import BOOTSTRAP_MARKUP_HIGHLIGHTER, \
    RESULTS_FRAGMENTER
from flask_sqlalchemy import SQLAlchemy
from jsonrpcserver import method
from marshmallow import Schema, fields
from whoosh.qparser import DisMaxParser
from whoosh.searching import ResultsPage

from labster.di import injector
from labster.domain2.model.demande import Demande
from labster.domain.models.faq import FaqEntry
from labster.extensions import whoosh
from labster.rbac import has_read_access, is_membre_dri
from labster.security import get_current_profile
from labster.types import JSONDict

PAGE_SIZE = 25

db = injector.get(SQLAlchemy)


@method
def search_api(q: str, page=1) -> JSONDict:
    q = q.strip()

    if q:
        results = search_results(q, page)
        demandes, faqs = results["demandes"], results["faqs"]
    else:
        demandes = []
        faqs = []

    ctx = {
        "q": q,
        "demandes": DemandeSchema().dump(demandes, many=True).data,
        "faqs": FaqEntrySchema().dump(faqs, many=True).data,
    }
    return ctx


def search_results(q, page):
    search_kwargs = {"limit": page * PAGE_SIZE}
    results = search(q, **search_kwargs)

    results.formatter = BOOTSTRAP_MARKUP_HIGHLIGHTER
    results.fragmenter = RESULTS_FRAGMENTER
    results = ResultsPage(results, page, PAGE_SIZE)

    def is_demande(hit):
        return hit["cls"] == "Demande"

    groups = toolz.groupby(is_demande, results)

    demande_ids = [hit["id"] for hit in groups.get(True, [])]
    faq_ids = [hit["id"] for hit in groups.get(False, [])]

    demandes = db.session.query(Demande).filter(Demande.id.in_(demande_ids)).all()
    demandes = filter_demandes_by_visibility(demandes)

    faqs = FaqEntry.query.filter(FaqEntry.id.in_(faq_ids)).all()

    return {"demandes": demandes, "faqs": faqs}


class FaqEntrySchema(Schema):
    id = fields.String()
    title = fields.String()
    # body = fields.String()
    # category = fields.String()


class StructureSchema(Schema):
    id = fields.String()
    nom = fields.String()


class UserSchema(Schema):
    id = fields.String()
    full_name = fields.String()


class DemandeSchema(Schema):
    id = fields.String()
    nom = fields.String()
    created_at = fields.DateTime()
    type = fields.String()
    wf_state = fields.String()
    icon_class = fields.String()

    porteur = fields.Nested(UserSchema)
    gestionnaire = fields.Nested(UserSchema)
    structure = fields.Nested(StructureSchema)


def filter_demandes_by_visibility(demandes):
    return [demande for demande in demandes if has_read_access(demande)]


def search(q, **search_args):
    """Interface to search indexes.

    :param q: unparsed search string.
    :param search_args: any valid parameter for
        :meth:`whoosh.searching.Search.search`. This includes `limit`,
        `groupedby` and `sortedby`
    """
    index = whoosh.index

    fields = {"name": 1.5, "text": 1.0}
    parser = DisMaxParser(fields, index.schema)
    query = parser.parse(q)

    # security access filter
    user = get_current_profile()

    if not is_membre_dri(user):
        pass
        # TODO
        # roles = {f"user:{user.id}", "all"}
        # for role in user.get_roles():
        #     if role.type in [RoleType.DIRECTION.value, RoleType.GDL.value]:
        #         structure = role.context
        #         structures = [structure] + structure.descendants()
        #         roles |= {f"org:{s.id}" for s in structures}
        #
        # terms = [wq.Term("allowed_roles_and_users", role) for role in roles]
        # query &= wq.Or(terms)

    with index.searcher(closereader=False) as searcher:
        # 'closereader' is needed, else results cannot by used outside 'with'
        # statement
        return searcher.search(query, **search_args)



from __future__ import annotations

import unicodedata
from functools import singledispatch
from typing import Iterable

import flask

from labster.domain2.model.demande import Demande
from labster.domain2.model.profile import Profile
from labster.domain2.model.structure import Structure
from labster.domain.models.demandes import Demande as OldDemande
from labster.domain.models.profiles import Profile as OldProfile


@singledispatch
def url_for(obj, **kw) -> str:
    raise RuntimeError(f"Unknown type ({type(obj)}) for obj {obj}")


@url_for.register(str)
def url_for_str(obj: str, **kw) -> str:
    return flask.url_for(obj, **kw)


@url_for.register(Profile)
def url_for_profile(profile: Profile, **kw) -> str:
    return flask.url_for("main.home", _anchor=f"/annuaire/users/{profile.id}", **kw)


@url_for.register(Structure)
def url_for_structure(structure: Structure, **kw) -> str:
    return flask.url_for(
        "main.home", _anchor=f"/annuaire/structures/{structure.id}", **kw
    )


@url_for.register(Demande)
def url_for_demande(demande: Demande, **kw) -> str:
    return flask.url_for("main.home", _anchor=f"/demandes/{demande.id}", **kw)


#
# Old stuff. Remove when ready.
#
@url_for.register(OldDemande)
def url_for_old_demande(demande: OldDemande, **kw) -> str:
    return flask.url_for("main.home", _anchor=f"/demandes/{demande.id}", **kw)


@url_for.register(OldProfile)
def url_for_old_profile(profile: OldProfile, **kw) -> str:
    return flask.url_for("main.home", _anchor=f"/annuaire/users/{profile.id}", **kw)


# @url_for.register(OrgUnit)
# def url_for_org_unit(org_unit: OrgUnit, **kw) -> str:
#     return flask.url_for(
#         "main.home", _anchor=f"/annuaire/structures/{org_unit.id}", **kw
#     )


def sort_by_name(iterable: Iterable) -> list:
    result = list(iterable)
    if not result:
        return []
    if hasattr(result[0], "prenom"):
        return sorted(result, key=lambda x: (x.nom, x.prenom))
    else:
        return sorted(result, key=lambda x: x.name)


def strip_accents(text: str) -> str:
    return unicodedata.normalize("NFD", text).encode("ascii", "ignore").decode("utf-8")



""""""
from __future__ import annotations

import datetime
from decimal import Decimal, ExtendedContext, localcontext

from babel.numbers import format_currency, format_decimal
from flask_babel import get_locale

from labster.domain2.services.constants import get_constant, get_constants

_PREC_1 = Decimal(".1")
_PREC_2 = Decimal(".01")

# Constantes
LOCALISATION_TO_IR = {
    "": Decimal(0),
    "Paris ou région parisienne": Decimal("0.03"),
    "Banyuls-sur-mer": Decimal(0),
    "Roscoff": Decimal(0),
    "Villefranche-sur-mer": Decimal("0.01"),
}


def POINT_INDICE() -> Decimal:
    try:
        return Decimal(get_constant("point_indice"))
    except RuntimeError:
        # For unit tests
        return Decimal("4.630291")


def get_ctx_for_demande(demande):
    try:
        locale = get_locale()
    except RuntimeError:
        locale = "fr"
    constants = get_constants()
    cst = constants["recrutement"]
    ctx = {
        "demande": demande,
        "today": datetime.date.today(),
        "contact": demande.gestionnaire or demande.porteur,
    }

    def format_eur(amount):
        return format_currency(amount, "EUR", locale=locale)

    with localcontext(ExtendedContext):
        salaire_brut_mensuel = demande.data.get("salaire_brut_mensuel", "")
        if salaire_brut_mensuel:
            salaire_brut = ctx["salaire_brut"] = Decimal(
                str_to_float(salaire_brut_mensuel)
            )
        else:
            salaire_brut = ctx["salaire_brut"] = Decimal("0")

        if demande.quotite_de_travail and demande.quotite_de_travail.endswith("%"):
            ctx["quotite_de_travail"] = str_to_int(demande.quotite_de_travail[0:-1])
        else:
            ctx["quotite_de_travail"] = 100

        # Cas particulier des 80% qui comptent pour 85.7%.
        if ctx["quotite_de_travail"] != 80:
            salaire_brut_etp = salaire_brut * 100 / ctx["quotite_de_travail"]
        else:
            salaire_brut_etp = salaire_brut * 1000 / 857

        ctx["salaire_brut_etp"] = Decimal(salaire_brut_etp)

        ctx["indice_majore"] = int(round(ctx["salaire_brut_etp"] / POINT_INDICE()))

        # if Doctorant, do not show indice_majore nor ir (thus ir to 0 for the rest).
        nature_du_recrutement = demande.data.get("nature_du_recrutement")
        ctx["is_doctorant"] = False
        if nature_du_recrutement == "Doctorant":
            ctx["ir"] = 0
            ctx["is_doctorant"] = True
        else:
            ctx["ir"] = LOCALISATION_TO_IR.get(demande.localisation, 0) * salaire_brut

        total_salaire_brut = ctx["total_salaire_brut"] = ctx["salaire_brut"] + ctx["ir"]

        if demande.indemnite_transport_en_commun == "oui":
            ctx["cout_transport"] = Decimal(cst["transport"])
        else:
            ctx["cout_transport"] = 0

        nombre_enfants_a_charge = str_to_int(demande.nombre_enfants_a_charge or 0)
        ctx["sft"] = sft(ctx["indice_majore"], nombre_enfants_a_charge)

        # Si la durée du contrat est supérieure ou égale à 12 mois =36%
        # Si la durée du contrat est inférieure à 12 mois = 37,5%
        # TODO: calculer la durée en mois
        if demande.duree_jours >= 365:
            coef_charges = Decimal(cst["charges_plus_12_mois"])
        else:
            coef_charges = Decimal(cst["charges_moins_12_mois"])
        coef_charges = coef_charges / 100

        ctx["coef_charges"] = coef_charges * 100
        ctx["montant_charges"] = (ctx["total_salaire_brut"] + ctx["sft"]) * coef_charges

        ctx["cout_mens_charges_comprises"] = (
            ctx["total_salaire_brut"]
            + ctx["cout_transport"]
            + ctx["sft"]
            + ctx["montant_charges"]
        )

        provision_risque_charge_employeur = cst["provision_risque_charge_employeur"]
        ctx["pourcentage_provision"] = provision_risque_charge_employeur
        ctx["provision_risque"] = ctx["cout_mens_charges_comprises"] * Decimal(
            provision_risque_charge_employeur / 100
        )
        ctx["cout_total"] = ctx["cout_mens_charges_comprises"] + ctx["provision_risque"]

        if salaire_brut:
            ctx["cout_total_proportion"] = ctx["cout_total"] / salaire_brut
        else:
            ctx["cout_total_proportion"] = Decimal(0)

        # cout periode
        date_debut = demande.date_debut
        date_fin = demande.date_fin
        cout_periode = ctx["cout_periode"] = []

        if date_debut and date_fin and date_fin > date_debut:
            ctx["nb_jours"] = (date_fin - date_debut).days + 1

            ctx["annees_cout_periode"] = range(date_debut.year, date_debut.year + 4)
            start = date_debut

            for year in range(date_debut.year, date_fin.year):
                end = start.replace(month=12, day=31)
                months = month_duration(start, end)
                cout_periode.append(
                    {
                        "months": months,
                        "brut": (total_salaire_brut * months),
                        "provision": (ctx["provision_risque"] * months),
                        "total": (ctx["cout_total"] * months),
                    }
                )
                start = datetime.date(year + 1, 1, 1)

            months = month_duration(start, date_fin)

            cout_periode.append(
                {
                    "months": months,
                    "brut": (total_salaire_brut * months),
                    "provision": (ctx["provision_risque"] * months),
                    "total": (ctx["cout_total"] * months),
                }
            )

            brut = sum(d["brut"] for d in cout_periode)
            provision = sum(d["provision"] for d in cout_periode)
            total = sum(d["total"] for d in cout_periode)
            total_months = sum(d["months"] for d in cout_periode)
            cp_total = {
                "months": format_decimal(total_months, locale=locale),
                "brut": format_eur(brut),
                "provision": format_eur(provision),
                "total": format_eur(total),
            }

            # round total to upper 500 unit
            total_arrondi = int(total)
            remainder = total_arrondi % 500
            if remainder:
                total_arrondi += 500 - remainder
            ctx["total_arrondi"] = format_eur(total_arrondi)

            # format values
            cp_total["total"] = format_eur(total)
            for d in cout_periode:
                d["months"] = format_decimal(d["months"], format="#0.#", locale=locale)
                d["brut"] = format_eur(d["brut"])
                d["provision"] = format_eur(d["provision"])
                d["total"] = format_eur(d["total"])

            # fill columns if needed
            while len(cout_periode) < 4:
                cout_periode.append({})
            cout_periode.append(cp_total)

        else:
            ctx["total_arrondi"] = Decimal(0)
            ctx["annees_cout_periode"] = []

    return ctx


def sft(indice_majore: int, nb_enfants_a_charge: int) -> Decimal:
    if indice_majore < 445:
        indice_base = 444
    elif indice_majore < 717:
        indice_base = indice_majore
    else:
        indice_base = 717

    salaire_base = indice_base * POINT_INDICE()

    if nb_enfants_a_charge == 0:
        result = Decimal(0)
    elif nb_enfants_a_charge == 1:
        result = Decimal("2.29")
    elif nb_enfants_a_charge == 2:
        result = Decimal("10.67") + Decimal("0.03") * salaire_base
    elif nb_enfants_a_charge == 3:
        result = Decimal("15.24") + Decimal("0.08") * salaire_base
    else:
        n = nb_enfants_a_charge - 3
        result = (
            Decimal("15.24")
            + n * Decimal("4.57")
            + (Decimal("0.08") + n * Decimal("0.06")) * salaire_base
        )
    return result.quantize(_PREC_2)


def cout_total_charge(demande):
    ctx = get_ctx_for_demande(demande)
    return ctx["total_arrondi"]


def days_in_month(dt):
    """Return the number of days in the month."""
    month = (dt.month % 12) + 1
    year = dt.year
    if month == 1:
        year += 1
    return (dt.replace(year=year, month=month, day=1) - datetime.timedelta(1)).day


def month_duration(start, end):
    """Duration in months with fractionnal part between 2 dates in same
    year."""
    assert start.year == end.year
    assert start <= end

    start_dim = Decimal(days_in_month(start))
    start_days = Decimal(1 + start_dim - start.day)

    if start.month == end.month:
        # special case
        days = Decimal(1 + end.day - start.day)
        return (days / start_dim).quantize(_PREC_1)

    months = start_days / start_dim

    start = start.replace(month=start.month + 1, day=1)
    months += end.month - start.month
    months += Decimal(end.day) / Decimal(days_in_month(end))

    return months.quantize(_PREC_1)


def str_to_int(s, default=0):
    if isinstance(s, int):
        return s
    if isinstance(s, float):
        return int(s)

    s = s or ""
    s = s.replace(" ", "")
    try:
        return int(s)
    except ValueError:
        return default


def str_to_float(s, default=0.0):
    if isinstance(s, float):
        return s
    if isinstance(s, int):
        return float(s)

    s = s or ""
    s = s.replace(" ", "")
    s = s.replace(",", ".")
    try:
        return float(s)
    except ValueError:
        return default
